import os

import numpy as np
import torch
from torch.distributions import Uniform
from tqdm import tqdm, trange

from .distributions import IndependentNormal, init_independent_normal
from .mcmc_base import AbstractMCMC, adapt_stepsize_dec, increment_steps


def grad_energy(point, target, x=None):
    point = point.detach().requires_grad_()
    if x is not None:
        energy = -target.log_prob(z=point, x=x)
    else:
        energy = -target.log_prob(point)

    grad = torch.autograd.grad(energy.sum(), point)[0]
    return energy, grad


def gan_energy(
    z,
    generator,
    discriminator,
    proposal,
    normalize_to_0_1,
    alpha=1.0,
    log_prob=False,
    z_transform=None,
):
    if z_transform is None:
        generator_points = generator(z)
    else:
        generator_points = generator(z_transform(z))
    if normalize_to_0_1:
        gan_part = -discriminator(generator_points).view(-1)
    else:
        sigmoid_gan_part = discriminator(generator_points)
        gan_part = -(torch.log(sigmoid_gan_part) - torch.log1p(-sigmoid_gan_part)).view(
            -1
        )

    # very strangely last vector in batch produces wrong gradient !
    try:
        proposal_part = -proposal.log_prob(z)
    except Exception:
        proposal_part = 0  # print(gan_part, z)

    energy = alpha * gan_part + proposal_part

    if not log_prob:
        return energy
    else:
        return -energy


def gan_energy_stylegan2_ada(
    z,
    generator,
    discriminator,
    proposal,
    normalize_to_0_1,
    log_prob=False,
    z_transform=None,
):
    label = torch.zeros([z.shape[0], generator.c_dim], device=z.device)
    if z_transform is None:
        generator_points = generator(z, label)
    else:
        generator_points = generator(z_transform(z), label)
    if normalize_to_0_1:
        gan_part = -discriminator(generator_points, label).view(-1)
    else:
        sigmoid_gan_part = discriminator(generator_points, label)
        gan_part = -(torch.log(sigmoid_gan_part) - torch.log1p(-sigmoid_gan_part)).view(
            -1
        )

    proposal_part = -proposal.log_prob(z)

    energy = gan_part + proposal_part
    if not log_prob:
        return energy
    else:
        return -energy


def sampling_f(
    dynamics,
    target,
    proposal,
    batch_size,
    n,
    path_to_save=None,
    file_name=None,
    every_step=None,
    continue_z=None,
    *args,
    **kwargs,
):
    z_last = []
    zs = []
    z = None
    for i in tqdm(range(0, n, batch_size)):
        if continue_z is None:
            z = proposal.sample([batch_size])
        else:
            j = i // batch_size
            z = continue_z[j][-1].clone().detach()
        z.requires_grad_(True)
        out = dynamics(z, target, proposal, *args, **kwargs)
        if isinstance(out, tuple):
            z_sp = out[0]
        else:
            z_sp = out
        last = z_sp[-1].data.cpu().numpy()
        zs_append = np.stack([o.data.cpu().numpy() for o in z_sp], axis=0)
        z_last.append(last)
        zs.append(zs_append)
        if (
            (file_name is not None)
            and (path_to_save is not None)
            and (every_step is not None)
        ):
            cur_file_name = file_name + f"_batch_num_{i}.npy"
            cur_path_to_save = os.path.join(path_to_save, cur_file_name)
            save_np_file = zs_append[::every_step, :, :]
            np.save(cur_path_to_save, save_np_file)
            print(
                f"file {cur_path_to_save} was saved, file shape = {save_np_file.shape}",
            )

    z_last_np = np.asarray(z_last).reshape(-1, z.shape[-1])
    zs = np.stack(zs, axis=0)
    return z_last_np, zs


def load_data_from_batches(n, batch_size, path_to_save, path_to_batches):
    load_np = []

    for i in tqdm(range(0, n, batch_size)):
        cur_file_name = path_to_batches + f"_batch_num_{i}.npy"
        cur_path_to_save = os.path.join(path_to_save, cur_file_name)
        cur_zs = np.load(cur_path_to_save)
        load_np.append(cur_zs)

    load_np = np.array(load_np)
    load_np = np.concatenate(
        load_np.transpose((0, 2, 1, 3)),
        axis=0,
    ).transpose((1, 0, 2))
    return load_np


def sampling_from_dynamics(dynamics):
    def sampling_dynamics_f(
        target,
        proposal,
        batch_size,
        n,
        path_to_save=None,
        file_name=None,
        every_step=None,
        continue_z=None,
        *args,
        **kwargs,
    ):
        return sampling_f(
            dynamics,
            target,
            proposal,
            batch_size,
            n,
            path_to_save,
            file_name,
            every_step,
            continue_z,
            *args,
            **kwargs,
        )

    return sampling_dynamics_f


def aggregate_sampling_output(z):
    return np.concatenate(z.transpose((0, 2, 1, 3)), axis=0).transpose(
        (1, 0, 2),
    )


def langevin_dynamics(
    z, target, proposal, n_steps, grad_step, eps_scale, verbose=False
):
    z_sp = []
    batch_size, _ = z.shape[0], z.shape[1]

    range_gen = trange if verbose else range

    for _ in range_gen(n_steps):
        z_sp.append(z)
        eps = eps_scale * proposal.sample([batch_size])

        E, grad = grad_energy(z, target, x=None)
        z = z - grad_step * grad + eps
        z = z.data
        z.requires_grad_(True)
    z_sp.append(z)
    return z_sp


langevin_sampling = sampling_from_dynamics(langevin_dynamics)


class ULA(AbstractMCMC):
    def __init__(self, **kwargs):
        super().__init__()
        self.device = kwargs.get("device", "cpu")
        self.grad_step = kwargs.get("grad_step", 1e-2)
        self.noise_scale = kwargs.get(
            "noise_scale",
            (2 * self.grad_step) ** 0.5,
        )
        self.verbose = kwargs.get("verbose", True)

    def __call__(
        self,
        z,
        target,
        proposal,
        n_steps,
        grad_step=None,
        noise_scale=None,
        verbose=None,
    ):
        grad_step = grad_step if grad_step is not None else self.grad_step
        noise_scale = noise_scale if noise_scale is not None else self.noise_scale
        verbose = verbose if verbose is None else self.verbose
        return langevin_dynamics(
            z, target, proposal, n_steps, grad_step, noise_scale, verbose
        )


def heuristics_step_size(
    mean_acceptance,
    target_acceptance,
    stepsize,
    factor=1.05,
    tol=0.03,
):
    if mean_acceptance - target_acceptance > tol:
        return stepsize * factor
    if target_acceptance - mean_acceptance > tol:
        return stepsize / factor
    return stepsize


def mala_dynamics(
    z,
    target,
    proposal,
    n_steps,
    grad_step,
    eps_scale,
    acceptance_rule="Hastings",
    adapt_stepsize=False,
):
    z_sp = [z.clone().detach()]
    batch_size, z_dim = z.shape[0], z.shape[1]
    device = z.device

    # !!
    std_norm = torch.distributions.multivariate_normal.MultivariateNormal(
        torch.zeros(z_dim).to(device),
        torch.eye(z_dim).to(device),
    )

    if adapt_stepsize:
        eps_scale = (2 * grad_step) ** (1 / 2)
    uniform = Uniform(low=0.0, high=1.0)
    acceptence = torch.zeros(batch_size).to(device)

    for _ in range(n_steps):
        eps = eps_scale * std_norm.sample([batch_size])

        E, grad = grad_energy(z, target, x=None)

        new_z = z - grad_step * grad + eps
        new_z = new_z.data
        new_z.requires_grad_(True)

        E_new, grad_new = grad_energy(new_z, target, x=None)

        energy_part = E - E_new

        propose_vec_1 = z - new_z + grad_step * grad_new
        propose_vec_2 = new_z - z + grad_step * grad

        # propose_part_1 = proposal.log_prob(propose_vec_1/eps_scale)
        # propose_part_2 = proposal.log_prob(propose_vec_2/eps_scale)
        propose_part_1 = std_norm.log_prob(propose_vec_1 / eps_scale)
        propose_part_2 = std_norm.log_prob(propose_vec_2 / eps_scale)

        propose_part = propose_part_1 - propose_part_2

        log_accept_prob = torch.zeros_like(propose_part)

        if acceptance_rule == "Hastings":
            log_accept_prob = propose_part + energy_part

        elif acceptance_rule == "Barker":
            log_ratio = propose_part + energy_part
            log_accept_prob = -torch.log(1.0 + torch.exp(-log_ratio))

        generate_uniform_var = uniform.sample([batch_size]).to(z.device)
        log_generate_uniform_var = torch.log(generate_uniform_var)
        mask = log_generate_uniform_var < log_accept_prob

        # Adapting heuristics
        if adapt_stepsize:
            mean_acceptance = mask.float().mean()
            grad_step = heuristics_step_size(
                mean_acceptance,
                target_acceptance=0.45,
                stepsize=grad_step,
            )
            eps_scale = (2 * grad_step) ** (1 / 2)
        ########

        acceptence += mask
        with torch.no_grad():
            z[mask] = new_z[mask].detach().clone()
            z = z.data
            z.requires_grad_(True)
            z_sp.append(z.clone().detach())

    return z_sp, acceptence


mala_sampling = sampling_from_dynamics(mala_dynamics)


class MALA(AbstractMCMC):
    def __init__(self, dim, **kwargs):
        super().__init__()
        self.dim = dim
        self.device = kwargs.get("device", "cpu")
        self.grad_step = kwargs.get("grad_step", 1e-2)
        self.noise_scale = kwargs.get(
            "noise_scale",
            (2 * self.grad_step) ** 0.5,
        )
        self.adapt_stepsize = kwargs.get("adapt_stepsize", False)
        self.mala_transition = MALATransition(dim, self.device)
        if self.adapt_stepsize:
            self.mala_transition.adapt_grad_step = self.grad_step
            self.mala_transition.adapt_sigma = self.noise_scale
        self.verbose = kwargs.get("verbose", True)
        self.beta = kwargs.get("beta", 1.0)  # inverse temperature

    # @increment_steps
    # @adapt_stepsize_dec
    def __call__(
        self,
        z,
        target,
        proposal,
        n_steps,
        grad_step=None,
        noise_scale=None,
        beta=None,
        adapt_stepsize=None,
        verbose=None,
    ):
        adapt_stepsize = (
            self.adapt_stepsize if adapt_stepsize is None else adapt_stepsize
        )
        grad_step = self.grad_step if grad_step is None else grad_step
        noise_scale = self.noise_scale if noise_scale is None else noise_scale
        verbose = self.verbose if verbose is None else verbose
        beta = self.beta if beta is None else beta

        zs = [z.clone().detach()]
        batch_size = z.shape[0]
        acceptance = torch.zeros(batch_size)

        range_gen = trange if verbose else range

        for _ in range_gen(n_steps):
            energy, grad = grad_energy(z, target)
            z, _, _, mask = self.mala_transition(
                z,
                energy,
                grad,
                grad_step=grad_step,
                sigma=noise_scale,
                beta=beta,
                target=target,
                adapt_stepsize=adapt_stepsize,
            )
            zs.append(z.detach().clone())
            acceptance += mask.float() / n_steps
        if adapt_stepsize:
            self.grad_step = self.mala_transition.adapt_grad_step

        return zs, acceptance


def mh_dynamics_normal_proposal(
    z,
    target,
    proposal,
    n_steps,
    eps_scale,
    acceptance_rule="Hastings",
):
    z_sp = [z.clone().detach()]
    batch_size, z_dim = z.shape[0], z.shape[1]
    device = z.device

    # !!
    std_norm = torch.distributions.multivariate_normal.MultivariateNormal(
        torch.zeros(z_dim).to(device),
        torch.eye(z_dim).to(device),
    )

    uniform = Uniform(low=0.0, high=1.0)
    acceptence = torch.zeros(batch_size).to(device)
    log_accept_prob = None

    for _ in range(n_steps):
        E = -target(z)

        eps = eps_scale * std_norm.sample([batch_size]).to(device)
        new_z = z + eps
        new_z = new_z.data

        E_new = -target(new_z)

        energy_part = E - E_new

        propose_vec_1 = z - new_z
        propose_vec_2 = new_z - z

        #    propose_part_1 = proposal.log_prob(propose_vec_1/eps_scale)
        #    propose_part_2 = proposal.log_prob(propose_vec_2/eps_scale)
        propose_part_1 = std_norm.log_prob(propose_vec_1 / eps_scale)
        propose_part_2 = std_norm.log_prob(propose_vec_2 / eps_scale)

        propose_part = propose_part_1 - propose_part_2

        if acceptance_rule == "Hastings":
            log_accept_prob = propose_part + energy_part

        elif acceptance_rule == "Barker":
            log_ratio = propose_part + energy_part
            log_accept_prob = -torch.log(1.0 + torch.exp(-log_ratio))

        generate_uniform_var = uniform.sample([batch_size]).to(z.device)
        log_generate_uniform_var = torch.log(generate_uniform_var)
        mask = log_generate_uniform_var < log_accept_prob

        acceptence += mask
        with torch.no_grad():
            z[mask] = new_z[mask].detach().clone()
            z = z.data
            z_sp.append(z.clone().detach())

    return z_sp, acceptence


mh_sampling_normal_proposal = sampling_from_dynamics(
    mh_dynamics_normal_proposal,
)


def xtry_langevin_dynamics(
    y,
    target,
    proposal,
    n_steps,
    grad_step,
    eps_scale,
    N,
):
    y_arr = [y.detach().clone()]

    batch_size, z_dim = y.shape[0], y.shape[1]

    for _ in range(n_steps):
        U = torch.randint(0, N, (batch_size,)).tolist()
        # X = y.unsqueeze(1).repeat(1, N, 1)
        noise = proposal.sample([batch_size, N])

        _, grad_y = grad_energy(y, target, x=None)

        g = y - grad_step * grad_y
        g = g.data

        g_N = g.unsqueeze(1).repeat(1, N, 1)
        X = g_N + eps_scale * noise
        X[np.arange(batch_size), U, :] = y

        X_batch = (
            torch.transpose(X, 0, 1).reshape((batch_size * N, z_dim)).detach().clone()
        )

        X_batch.requires_grad_(True)
        minus_log_pi_x, minus_g_log_pi_x = grad_energy(X_batch, target, x=None)
        # e_grad(X_batch, normal, gen, dis, alpha, e_batch=True, ret_e=True)
        T_x = X_batch - grad_step * minus_g_log_pi_x
        T_x = torch.transpose(
            T_x.reshape(list(X.shape[:-1][::-1]) + [X.shape[-1]]),
            0,
            1,
        )
        minus_log_pi_x = minus_log_pi_x.reshape(X.shape[:-1][::-1]).T

        log_gauss = -(X[:, :, None] - T_x[:, None, :]).norm(
            p=2,
            dim=-1,
        ) ** 2 / (2 * eps_scale**2)
        sum_log_gauss = log_gauss.sum(1) - torch.diagonal(
            log_gauss,
            dim1=1,
            dim2=2,
        )

        log_weight = -minus_log_pi_x + sum_log_gauss

        max_logs = torch.max(log_weight, dim=1)[0].unsqueeze(-1).repeat((1, N))
        log_weight = log_weight - max_logs
        weight = torch.exp(log_weight)
        sum_weight = torch.sum(weight, dim=1)
        weight = weight / sum_weight[:, None]

        weight[weight != weight] = 0.0
        weight[weight.sum(1) == 0.0] = 1.0

        indices = torch.multinomial(weight, 1).squeeze().tolist()

        y = X[np.arange(batch_size), indices, :]
        y = y.data
        y.requires_grad_(True)
        y_arr.append(y.detach().clone())

    return y_arr


class MHKernel:
    @staticmethod
    def log_prob(
        log_pi1,
        log_pi2,
        log_transition_forward,
        log_transition_backward,
    ):
        return (log_pi2 + log_transition_backward) - (log_pi1 + log_transition_forward)

    @staticmethod
    def get_mask(
        log_pi1,
        log_pi2,
        log_transition_forward,
        log_transition_backward,
    ):
        batch_size = log_pi1.shape[0]
        device = log_pi1.device
        acc_log_prob = MHKernel.log_prob(
            log_pi1,
            log_pi2,
            log_transition_forward,
            log_transition_backward,
        )
        generate_uniform_var = torch.rand([batch_size], device=device)
        log_generate_uniform_var = torch.log(generate_uniform_var)
        mask = log_generate_uniform_var < acc_log_prob
        return mask


class MALATransition:
    def __init__(self, z_dim, device):
        self.device = device
        self.z_dim = z_dim

        loc = torch.zeros(z_dim).to(device)
        scale = torch.ones(z_dim).to(device)
        self.stand_normal = IndependentNormal(
            dim=z_dim,
            device=device,
            loc=loc,
            scale=scale,
        )

        self.adapt_grad_step = 0.0
        self.adapt_sigma = 0.0

    @staticmethod
    def get_mh_kernel_log_prob(
        log_pi1,
        log_pi2,
        log_transition_forward,
        log_transition_backward,
    ):
        return (log_pi2 + log_transition_backward) - (log_pi1 + log_transition_forward)

    def get_langevin_transition_kernel(
        self,
        z1,
        z2,
        grad,
        grad_step,
        sigma=None,
    ):
        if sigma is None:
            sigma = (2 * grad_step) ** 0.5
        loc = z2 - (z1 - grad_step * grad)
        log_prob = self.stand_normal.log_prob(loc / sigma)
        return log_prob

    def compute_log_probs(
        self,
        z,
        z_new,
        energy,
        grad,
        grad_step,
        sigma,
        target=None,
        beta=1.0,
    ):
        energy_new, grad_new = grad_energy(z_new, target, x=None)
        log_transition_forward = self.get_langevin_transition_kernel(
            z,
            z_new,
            grad,
            beta * grad_step,
            sigma,
        )
        log_transition_backward = self.get_langevin_transition_kernel(
            z_new,
            z,
            grad_new,
            beta * grad_step,
            sigma,
        )
        log_prob = MHKernel.log_prob(
            -beta * energy,
            -beta * energy_new,
            log_transition_forward,
            log_transition_backward,
        )

        return log_prob, energy_new, grad_new

    def get_mask(
        self,
        z,
        z_new,
        energy,
        grad,
        grad_step,
        sigma,
        target=None,
        beta=1.0,
    ):
        energy_new, grad_new = grad_energy(z_new, target, x=None)
        log_transition_forward = self.get_langevin_transition_kernel(
            z,
            z_new,
            grad,
            beta * grad_step,
            sigma,
        )
        log_transition_backward = self.get_langevin_transition_kernel(
            z_new,
            z,
            grad_new,
            beta * grad_step,
            sigma,
        )

        mask = MHKernel.get_mask(
            -beta * energy,
            -beta * energy_new,
            log_transition_forward,
            log_transition_backward,
        )
        return mask, energy_new, grad_new

    def do_transition_step(
        self,
        z,
        z_new,
        energy,
        grad,
        grad_step,
        sigma,
        target=None,
        beta=1.0,
        adapt_stepsize=False,
    ):
        # if adapt_stepsize is True:
        #     if self.adapt_grad_step != 0 and self.adapt_sigma != 0:
        #         grad_step = self.adapt_grad_step
        #         sigma = self.adapt_sigma

        # acc_log_prob, energy_new, grad_new = self.compute_log_probs(z, z_new,
        #                                                             energy,
        #                                                             grad,
        #                                                             grad_step,
        #                                                             sigma,
        #                                                             target,
        #                                                             beta)

        # generate_uniform_var = self.uniform.sample([z.shape[0]]).to(z.device)
        # log_generate_uniform_var = torch.log(generate_uniform_var)
        # mask = log_generate_uniform_var < acc_log_prob
        mask, energy_new, grad_new = self.get_mask(
            z,
            z_new,
            energy,
            grad,
            grad_step,
            sigma,
            target,
            beta,
        )

        with torch.no_grad():
            z[mask] = z_new[mask].detach().clone()
            z = z.data
            z.requires_grad_(True)
            energy = energy.float()
            energy[mask] = energy_new[mask].float()
            energy[~mask] = energy[~mask]
            grad[mask] = grad_new[mask]
            grad[~mask] = grad[~mask]

        if adapt_stepsize:
            mean_acceptance = mask.float().mean()
            self.adapt_grad_step = heuristics_step_size(
                mean_acceptance,
                target_acceptance=0.45,
                stepsize=grad_step,
            )
            self.adapt_sigma = (2 * self.adapt_grad_step) ** 0.5

        return z, energy, grad, mask

    def __call__(
        self,
        z,
        energy,
        grad,
        grad_step=None,
        sigma=None,
        target=None,
        beta=1.0,
        adapt_stepsize=False,
    ):
        if adapt_stepsize is True or grad_step is None:
            if self.adapt_grad_step != 0 and self.adapt_sigma != 0:
                grad_step = self.adapt_grad_step
                sigma = self.adapt_sigma

        z_new = (
            z - grad_step * beta * grad + sigma * self.stand_normal.sample(z.shape[:-1])
        )
        z_new, energy, grad, mask = self.do_transition_step(
            z,
            z_new,
            energy,
            grad,
            grad_step,
            sigma,
            target=target,
            beta=beta,
            adapt_stepsize=adapt_stepsize,
        )
        return z_new, energy, grad, mask


def tempered_transitions_dynamics(
    z,
    target,
    proposal,
    n_steps,
    grad_step,
    eps_scale,
    betas,
):
    z_sp = [z.clone().detach()]
    batch_size, z_dim = z.shape[0], z.shape[1]
    device = z.device

    mala_transition = MALATransition(z_dim, device)

    acceptence = torch.zeros(batch_size).to(device)

    betas = np.array(betas)
    betas_diff = torch.FloatTensor(betas[:-1] - betas[1:])
    E0, grad0 = grad_energy(z, target, x=None)
    for _ in range(n_steps):
        z_forward = z.clone().detach()
        z_forward.requires_grad_(True)
        energy_forward = torch.zeros(batch_size, len(betas) - 1)
        energy_forward[:, 0] = E0

        E = E0  # data #.detach().clone()
        grad = grad0  # data #detach().clone()

        for i, beta in enumerate(betas[1:-1]):
            eps = eps_scale * proposal.sample([batch_size])

            z_forward_new = z_forward - grad_step * beta * grad + eps
            (z_forward, E, grad, mask) = mala_transition.do_transition_step(
                z_forward,
                z_forward_new,
                E,
                grad,
                grad_step,
                eps_scale,
                target,
                beta=beta,
            )
            energy_forward[:, i + 1] = E

        z_backward = z_forward.clone().detach()
        z_backward.requires_grad_(True)
        energy_backward = torch.zeros(batch_size, len(betas) - 1)
        energy_backward[:, -1] = E
        for i, beta in enumerate(betas[::-1][1:-1]):
            eps = eps_scale * proposal.sample([batch_size])

            z_backward_new = z_backward - grad_step * beta * grad + eps
            (z_backward, E, grad, mask) = mala_transition.do_transition_step(
                z_backward,
                z_backward_new,
                E,
                grad,
                grad_step,
                eps_scale,
                target,
                beta=beta,
            )
            j = len(betas) - i - 3
            energy_backward[:, j] = E

        F_forward = (betas_diff * energy_forward).sum(-1)
        F_backward = (betas_diff * energy_backward).sum(-1)
        log_accept_prob = F_forward - F_backward

        generate_uniform_var = mala_transition.uniform.sample([batch_size]).to(
            z.device,
        )
        log_generate_uniform_var = torch.log(generate_uniform_var)
        mask = log_generate_uniform_var < log_accept_prob

        acceptence += mask
        with torch.no_grad():
            z[mask] = z_backward[mask].detach().clone()
            z = z.data
            z.requires_grad_(True)
            z_sp.append(z.clone().detach())

            E0[mask] = E[mask].data

            grad0[mask] = grad[mask].data
            grad0.requires_grad_(True)

    return z_sp, acceptence


tempered_transitions_sampling = sampling_from_dynamics(
    tempered_transitions_dynamics,
)


def i_ais_z_dynamics(z, target, n_steps, grad_step, eps_scale, N, betas):
    z_sp = [z.clone().detach()]
    batch_size, z_dim = z.shape[0], z.shape[1]
    device = z.device

    mala_transition = MALATransition(z_dim, device)
    acceptence_rate = 0.0

    betas = np.array(betas)
    betas_diff = torch.FloatTensor(betas[:-1] - betas[1:])  # n-1

    scale_proposal = 1.0
    proposal = init_independent_normal(scale_proposal, z_dim, device)

    E_last, grad_last = grad_energy(z, target, x=None)
    energy_backward_last = None
    for step in tqdm(range(n_steps)):
        if step == 0:
            N_ = N
        else:
            N_ = N - 1

        Z = z.unsqueeze(1).repeat(1, N_, 1)
        z_batch = (
            torch.transpose(Z, 0, 1).reshape((batch_size * N_, z_dim)).detach().clone()
        )
        z_batch.requires_grad_(True)

        E = E_last.unsqueeze(1).repeat(1, N_)
        grad = grad_last.unsqueeze(1).repeat(1, N_, 1)
        grad = torch.transpose(grad, 0, 1).reshape((batch_size * N_, z_dim)).data
        # detach().clone()

        z_backward = z_batch
        z_backward.requires_grad_(True)
        energy_backward = torch.zeros(batch_size, N_, len(betas) - 1)  # n-1
        energy_backward[..., len(betas) - 2] = E  # detach().clone()

        E = E.T.reshape(-1)
        # betas[0] = 1, betas[n-1] = 0, betas[::-1][1:-1] - increasing
        # length is n-2
        for i, beta in enumerate(betas[::-1][1:-1]):
            j = len(betas) - i - 3
            eps = eps_scale * proposal.sample([batch_size * N_])

            z_backward_new = z_backward - beta * grad_step * grad + eps
            (z_backward, E, grad, mask) = mala_transition.do_transition_step(
                z_backward,
                z_backward_new,
                E,
                grad,
                grad_step,
                eps_scale,
                target,
                beta=beta,
            )
            accept = mask.sum()
            acceptence_rate += accept
            E_ = E.reshape(Z.shape[:-1][::-1]).T
            energy_backward[..., j] = E_.detach().clone()

        z_backward = torch.transpose(
            z_backward.reshape(list(Z.shape[:-1][::-1]) + [Z.shape[-1]]),
            0,
            1,
        )
        grad = torch.transpose(
            grad.reshape(list(Z.shape[:-1][::-1]) + [Z.shape[-1]]),
            0,
            1,
        )
        # E = E.reshape(Z.shape[:-1][::-1]).T

        if step > 0:
            z_backward = torch.cat([z_backward, z.unsqueeze(1)], dim=1)
            energy_backward = torch.cat(
                [energy_backward, energy_backward_last.unsqueeze(1)],
                dim=1,
            )
            grad = torch.cat([grad, grad_last.unsqueeze(1)], dim=1)

        E = energy_backward[..., 0]

        F_backward = (betas_diff[None, None, :] * energy_backward).sum(-1)
        log_weights = -F_backward

        max_logs = torch.max(log_weights, dim=1)[0]  # .unsqueeze(-1).repeat((1, N))
        log_weights = log_weights - max_logs[:, None]
        sum_weights = torch.logsumexp(log_weights, dim=1)
        log_weights = log_weights - sum_weights[:, None]
        weights = log_weights.exp()
        weights[weights != weights] = 0.0
        weights[weights.sum(1) == 0.0] = 1.0

        indices = torch.multinomial(weights, 1).squeeze().tolist()

        z = z_backward[np.arange(batch_size), indices, :]
        z = z.data
        z.requires_grad_(True)
        z_sp.append(z.detach().clone())

        E_last = E[np.arange(batch_size), indices].data

        grad_last = grad[np.arange(batch_size), indices, :].data
        grad_last.requires_grad_(True)

        energy_backward_last = energy_backward[np.arange(batch_size), indices, :]

    acceptence_rate = (
        acceptence_rate / (batch_size * N) / len(betas[::-1][1:-1]) / n_steps
    )
    return z_sp, acceptence_rate


def i_ais_v_dynamics(z, target, n_steps, grad_step, eps_scale, N, betas, rho):
    z_sp = [z.clone().detach()]
    batch_size, z_dim = z.shape[0], z.shape[1]
    device = z.device

    # mala_transition = CiterMALATransition(z_dim, device)
    mala_transition = MALATransition(z_dim, device)
    acceptence_rate = 0.0

    betas = np.array(betas)
    betas_diff = torch.FloatTensor(betas[:-1] - betas[1:])  # n-1

    scale_proposal = 1.0
    proposal = init_independent_normal(scale_proposal, z_dim, device)

    z_last = z
    z_backward_last = None
    energy_backward_last = None
    for step in tqdm(range(n_steps)):
        if step == 0:
            N_ = N
        else:
            N_ = N - 1

        z = z_last.unsqueeze(1).repeat(1, N_, 1)

        z = rho * z + ((1 - rho**2) ** 0.5) * proposal.sample(
            [batch_size, N_],
        )

        z_batch = (
            torch.transpose(z, 0, 1).reshape((batch_size * N_, z_dim)).detach().clone()
        )
        z_batch.requires_grad_(True)

        # E = E.unsqueeze(1).repeat(1, N)
        # grad = grad.unsqueeze(1).repeat(1, N, 1)
        # grad = torch.transpose(grad, 0, 1).reshape((batch_size*N, z_dim)).data #detach().clone()

        E, grad = grad_energy(z_batch, target)
        E = E.reshape(N_, batch_size).T

        z_backward = z_batch
        z_backward.requires_grad_(True)
        energy_backward = torch.zeros(batch_size, N_, len(betas) - 1)  # n-1
        energy_backward[..., len(betas) - 2] = E  # detach().clone()

        E = E.T.reshape(-1)
        # betas[0] = 1, betas[n-1] = 0, betas[::-1][1:-1] - increasing
        # length is n-2
        for i, beta in enumerate(betas[::-1][1:-1]):
            j = len(betas) - i - 3
            eps = eps_scale * proposal.sample([batch_size * N_])

            # z_backward_new = (1. - grad_step) * z_backward - beta * grad_step * grad + eps
            z_backward_new = z_backward - beta * grad_step * grad + eps
            (z_backward, E, grad, mask) = mala_transition.do_transition_step(
                z_backward,
                z_backward_new,
                E,
                grad,
                grad_step,
                eps_scale,
                target,
                beta=beta,
            )
            accept = mask.sum()
            acceptence_rate += accept
            E_ = E.reshape(z.shape[:-1][::-1]).T
            energy_backward[..., j] = E_.detach().clone()

        z_backward = torch.transpose(
            z_backward.reshape(list(z.shape[:-1][::-1]) + [z.shape[-1]]),
            0,
            1,
        )
        # grad = torch.transpose(grad.reshape(list(z.shape[:-1][::-1]) + [z.shape[-1]]), 0, 1)
        # E = E.reshape(z.shape[:-1][::-1]).T

        if step > 0:
            z = torch.cat([z, z_last.unsqueeze(1)], dim=1)
            z_backward = torch.cat(
                [z_backward, z_backward_last.unsqueeze(1)],
                dim=1,
            )
            energy_backward = torch.cat(
                [energy_backward, energy_backward_last.unsqueeze(1)],
                dim=1,
            )
            # grad = torch.cat([grad, grad_last.unsqueeze(1)], dim=1)

        # E = energy_backward[..., 0]

        F_backward = (betas_diff[None, None, :] * energy_backward).sum(-1)
        log_weights = -F_backward

        max_logs = torch.max(log_weights, dim=1)[0]
        log_weights = log_weights - max_logs[:, None]
        sum_weights = torch.logsumexp(log_weights, dim=1)
        log_weights = log_weights - sum_weights[:, None]
        weights = log_weights.exp()
        weights[weights != weights] = 0.0
        weights[weights.sum(1) == 0.0] = 1.0

        indices = torch.multinomial(weights, 1).squeeze().tolist()

        z_last = z[np.arange(batch_size), indices, :]
        z_last = z_last.data
        z_last.requires_grad_(True)

        z_backward_last = z_backward[np.arange(batch_size), indices, :]
        z_sp.append(z_backward_last.detach().clone())

        # grad_last = grad[np.arange(batch_size), indices, :].data
        # grad_last.requires_grad_(True)

        energy_backward_last = energy_backward[np.arange(batch_size), indices, :]

    acceptence_rate = 1.0
    # acceptence_rate/(batch_size*N)/len(betas[::-1][1:-1])/n_steps
    return z_sp, acceptence_rate


def i_ais_b_dynamics(z, target, n_steps, grad_step, eps_scale, N, betas, rho):
    z_sp = [z.clone().detach()]
    batch_size, z_dim = z.shape[0], z.shape[1]
    device = z.device

    mala_transition = MALATransition(z_dim, device)
    acceptence_rate = 0.0

    betas = np.array(betas)
    betas_diff = torch.FloatTensor(betas[:-1] - betas[1:])  # n-1

    scale_proposal = 1.0
    proposal = init_independent_normal(scale_proposal, z_dim, device)

    # E_last, grad_last = grad_energy(z, target, x=None)
    z_last = z
    z_backward_last = None
    energy_backward_last = None
    grad_last = None
    for step in tqdm(range(n_steps)):
        if step == 0:
            N_ = N
        else:
            N_ = N - 1

        z = z_last.unsqueeze(1).repeat(1, N_, 1)

        z = rho * z + ((1 - rho**2) ** 0.5) * proposal.sample(
            [batch_size, N_],
        )

        z_batch = (
            torch.transpose(z, 0, 1).reshape((batch_size * N_, z_dim)).detach().clone()
        )
        z_batch.requires_grad_(True)

        # E = E.unsqueeze(1).repeat(1, N)
        # grad = grad.unsqueeze(1).repeat(1, N, 1)
        # grad = torch.transpose(grad, 0, 1).reshape((batch_size*N, z_dim)).data #detach().clone()

        E, grad = grad_energy(z_batch, target)
        E = E.reshape(N_, batch_size).T

        z_backward = z_batch
        z_backward.requires_grad_(True)
        energy_backward = torch.zeros(batch_size, N_, len(betas) - 1)  # n-1
        energy_backward[..., len(betas) - 2] = E  # .detach().clone()

        E = E.T.reshape(-1)
        # betas[0] = 1, betas[n-1] = 0, betas[::-1][1:-1] - increasing
        # length is n-2
        for i, beta in enumerate(betas[::-1][1:-1]):
            j = len(betas) - i - 3
            eps = eps_scale * proposal.sample([batch_size * N_])

            # z_backward_new = (1 - grad_step) * z_backward - beta * grad_step * grad + eps
            z_backward_new = z_backward - beta * grad_step * grad + eps
            (z_backward, E, grad, mask) = mala_transition.do_transition_step(
                z_backward,
                z_backward_new,
                E,
                grad,
                grad_step,
                eps_scale,
                target,
                beta=beta,
            )
            accept = mask.sum()
            acceptence_rate += accept

            E_ = E.reshape(z.shape[:-1][::-1]).T
            energy_backward[..., j] = E_.detach().clone()

        z_backward = torch.transpose(
            z_backward.reshape(list(z.shape[:-1][::-1]) + [z.shape[-1]]),
            0,
            1,
        )
        grad = torch.transpose(
            grad.reshape(list(z.shape[:-1][::-1]) + [z.shape[-1]]),
            0,
            1,
        )
        # E = E.reshape(z.shape[:-1][::-1]).T

        if step > 0:
            # z = torch.cat([z, z_last.unsqueeze(1)], dim=1)
            z_backward = torch.cat(
                [z_backward, z_backward_last.unsqueeze(1)],
                dim=1,
            )
            energy_backward = torch.cat(
                [energy_backward, energy_backward_last.unsqueeze(1)],
                dim=1,
            )
            grad = torch.cat([grad, grad_last.unsqueeze(1)], dim=1)

        E = energy_backward[..., 0]

        F_backward = (betas_diff[None, None, :] * energy_backward).sum(-1)
        log_weights = -F_backward

        max_logs = torch.max(log_weights, dim=1)[0]
        log_weights = log_weights - max_logs[:, None]
        sum_weights = torch.logsumexp(log_weights, dim=1)
        log_weights = log_weights - sum_weights[:, None]
        weights = log_weights.exp()
        weights[weights != weights] = 0.0
        weights[weights.sum(1) == 0.0] = 1.0

        indices = torch.multinomial(weights, 1).squeeze().tolist()

        z_backward = z_backward[np.arange(batch_size), indices, :]
        z_backward = z_backward.data
        z_backward.requires_grad_(True)

        z_sp.append(z_backward.detach().clone())

        z_backward_last = z_backward
        energy_backward_last = energy_backward[np.arange(batch_size), indices, :]

        E = E[np.arange(batch_size), indices].data

        grad = grad[np.arange(batch_size), indices, :].data
        grad.requires_grad_(True)
        grad_last = grad

        z_backward = z_backward.detach().clone()
        z_backward.requires_grad_(True)

        for i, beta in enumerate(betas[:-1]):
            eps = eps_scale * proposal.sample([batch_size])

            # z_backward_new = (1 - grad_step) * z_backward - beta * grad_step * grad + eps
            z_backward_new = z_backward - beta * grad_step * grad + eps
            (z_backward, E, grad, mask) = mala_transition.do_transition_step(
                z_backward,
                z_backward_new,
                E,
                grad,
                grad_step,
                eps_scale,
                target,
                beta=beta,
            )

        z = z_backward.data
        z.requires_grad_(True)
        z_last = z
        # z_sp.append(z.detach().clone())
    acceptence_rate = (
        acceptence_rate / (batch_size * N) / len(betas[::-1][1:-1]) / n_steps
    )
    return z_sp, acceptence_rate


def compute_probs_from_log_probs(log_probs):
    mask_zeros = log_probs > 0.0
    log_probs[mask_zeros] = 0.0
    probs = log_probs.exp()
    return probs


def citerais_mala_dynamics(
    z,
    target,
    n_steps,
    grad_step,
    eps_scale,
    N,
    betas,
    rhos,
):
    z_sp = [z[:, -1, :].clone().detach()]
    batch_size, T, z_dim = z.shape[0], z.shape[1], z.shape[2]
    T = T - 1  # ??
    # T = len(betas) - 2
    device = z.device

    mala_transition = MALATransition(z_dim, device)

    betas = np.array(betas)
    betas_diff = torch.FloatTensor(betas[:-1] - betas[1:]).to(device)  # n-1

    z_flat = (
        torch.transpose(z, 0, 1)
        .reshape((batch_size * z.shape[1], z_dim))
        .detach()
        .clone()
    )
    z_flat.requires_grad_(True)
    E_flat, grad_flat = grad_energy(z_flat, target, None)
    grad = (
        torch.transpose(
            grad_flat.reshape(list(z.shape[:-1][::-1]) + [z.shape[-1]]),
            0,
            1,
        )
        .detach()
        .clone()
    )
    E = E_flat.reshape(z.shape[:-1][::-1]).T.data

    for _ in trange(n_steps):
        v = torch.zeros((batch_size, T + 1, N, z_dim), dtype=z.dtype).to(
            device,
        )
        u = torch.zeros((batch_size, T + 1, N), dtype=z.dtype).to(device)

        # step 1
        kappa = torch.zeros((batch_size, T + 1, z_dim), dtype=z.dtype).to(
            device,
        )
        kappa_t_noise = mala_transition.stand_normal.sample(
            [batch_size, T + 1],
        )
        kappa[:, 0, :] = (
            rhos[-1] * z[:, 0, :]
            + ((1 - rhos[-1] ** 2) ** 0.5) * kappa_t_noise[:, 0, :]
        )

        for t in range(1, T + 1):
            beta = betas[::-1][t]
            rho = rhos[::-1][t]

            not_equal_mask = torch.ne(z[:, t, :], z[:, t - 1, :]).max(dim=-1)[0]
            equal_mask = ~not_equal_mask
            num_not_equal = not_equal_mask.sum()
            num_equal = equal_mask.sum()

            if num_not_equal > 0:
                z_t_not_equal = z[not_equal_mask, t, :].detach().clone()
                z_t_1_not_equal = z[not_equal_mask, t - 1, :].detach().clone()
                z_t_not_equal.requires_grad_(True)
                z_t_1_not_equal.requires_grad_(True)

                # E_t_1_not_equal, grad_t_1_not_equal = grad_energy(z_t_1_not_equal, target, x=None)
                E_t_1_not_equal = E[not_equal_mask, t - 1]
                grad_t_1_not_equal = grad[not_equal_mask, t - 1, :]

                (log_probs, _, _) = mala_transition.compute_log_probs(
                    z_t_1_not_equal,
                    z_t_not_equal,
                    E_t_1_not_equal,
                    grad_t_1_not_equal,
                    grad_step,
                    eps_scale,
                    target,
                    beta=beta,
                )

                # v[not_equal_mask, t, 0, :] = (z_t_not_equal - (1. - grad_step) * z_t_1_not_equal
                #                               + grad_step * beta * grad_t_1_not_equal)/ (2 * grad_step)**.5
                #                               # eps_scale
                v[not_equal_mask, t, 0, :] = (
                    z_t_not_equal
                    - z_t_1_not_equal
                    + grad_step * beta * grad_t_1_not_equal
                ) / eps_scale

                probs = compute_probs_from_log_probs(log_probs)
                generate_uniform_var = mala_transition.uniform.sample(
                    [probs.shape[0]],
                ).to(probs.device)
                weight_uniform_var = generate_uniform_var * probs

                u[not_equal_mask, t, 0] = weight_uniform_var.float()

            if num_equal > 0:
                z_t_equal = z[equal_mask, t, :]
                z_t_1_equal = z[equal_mask, t - 1, :].detach().clone()
                z_t_1_equal.requires_grad_(True)

                # E_t_1_equal, grad_t_1_equal = grad_energy(z_t_1_equal,
                #                                           target, x=None)
                E_t_1_equal = E[equal_mask, t - 1].detach().clone()
                grad_t_1_equal = grad[equal_mask, t - 1, :].detach().clone()

                second_part_no_noise = (
                    1.0 - grad_step
                ) * z_t_1_equal - grad_step * beta * grad_t_1_equal
                # second_part_no_noise = z_t_1_equal - grad_step * beta * grad_t_1_equal

                stop = False
                # num_updates = 0
                update_mask = torch.zeros(num_equal, dtype=torch.bool).to(
                    z_t_equal.device,
                )
                z_t_1_equal = z_t_1_equal.detach().clone()
                z_t_1_equal.requires_grad_(True)

                while not stop:
                    cur_u = mala_transition.uniform.sample([num_equal]).to(
                        z_t_equal.device,
                    )
                    cur_v = mala_transition.stand_normal.sample(
                        [num_equal],
                    ).to(z_t_equal.device)
                    second_part = second_part_no_noise + cur_v * eps_scale
                    second_part = second_part.detach().clone()
                    second_part.requires_grad_(True)

                    (log_probs, _, _) = mala_transition.compute_log_probs(
                        z_t_1_equal,
                        second_part,
                        E_t_1_equal,
                        grad_t_1_equal,
                        grad_step,
                        eps_scale,
                        target,
                        beta=beta,
                    )

                    probs = compute_probs_from_log_probs(log_probs)
                    mask_assign = cur_u <= probs
                    new_assign = torch.logical_and(mask_assign, ~update_mask)

                    u[equal_mask, t, 0][new_assign] = cur_u[new_assign]
                    v[equal_mask, t, 0, :][new_assign] = cur_v[new_assign]

                    update_mask = torch.logical_or(update_mask, new_assign)
                    updates_num = update_mask.sum()
                    if updates_num == num_equal:
                        stop = True

            kappa[:, t, :] = (
                rho * v[:, t, 0, :] + ((1 - rho**2) ** 0.5) * kappa_t_noise[:, t, :]
            )

        # step 2
        # W = mala_transition.stand_normal.sample([batch_size, N - 1])
        # Z - tensor (bs, T + 1, N, dim)
        Z = torch.zeros((batch_size, T + 1, N, z_dim), dtype=z.dtype).to(
            device,
        )

        kappa_repeat = kappa[:, 0, :].unsqueeze(1).repeat(1, N - 1, 1)
        kappa_N_noise = mala_transition.stand_normal.sample(
            [batch_size, N - 1],
        )
        kappa_repeat_N = kappa.unsqueeze(2).repeat(1, 1, N - 1, 1)

        Z[:, :, 0, :] = z
        Z[:, 0, 1:, :] = (
            rhos[-1] * kappa_repeat + ((1 - rhos[-1] ** 2) ** 0.5) * kappa_N_noise
        )

        kappa_flat = (
            torch.transpose(Z[:, 0, 1:, :], 0, 1)
            .reshape((batch_size * (N - 1), z_dim))
            .detach()
            .clone()
        )
        kappa_E_flat, kappa_grad_flat = grad_energy(kappa_flat, target)
        kappa_E = kappa_E_flat.reshape(N - 1, batch_size).T
        kappa_grad = torch.transpose(
            kappa_grad_flat.reshape(N - 1, batch_size, z_dim),
            0,
            1,
        )

        energy = torch.zeros(batch_size, T + 1, N).to(device)

        energy[:, :, 0] = E.data
        energy[:, 0, 1:] = kappa_E

        grads = torch.zeros(batch_size, T + 1, N, z_dim).to(device)
        grads[:, :, 0, :] = grad.detach().clone()
        grads[:, 0, 1:, :] = kappa_grad

        W_2 = mala_transition.stand_normal.sample([batch_size, T, N - 1])

        for t in range(1, T + 1):
            beta = betas[::-1][t]
            rho = rhos[::-1][t]

            v[:, t, 1:, :] = (
                rho * kappa_repeat_N[:, t, :, :]
                + ((1 - rho**2) ** 0.5) * W_2[:, t - 1, :, :]
            )
            # z_t_1_j_shape = Z[:, t - 1, 1:, :].shape
            z_t_1_j_flatten = (
                torch.transpose(Z[:, t - 1, 1:, :], 0, 1)
                .reshape((batch_size * (N - 1), z_dim))
                .detach()
                .clone()
                .to(device)
            )
            z_t_1_j_flatten.requires_grad_(True)

            # _, grad_z_t_1_j_flatten = grad_energy(z_t_1_j_flatten, target, x=None)
            # grad_z_t_1_j = (torch.transpose(grad_z_t_1_j_flatten
            #                 .reshape(list(z_t_1_j_shape[:-1][::-1]) + [z_t_1_j_shape[-1]]), 0, 1))
            grad_z_t_1_j = grads[:, t - 1, 1:, :]
            grad_z_t_1_j_flatten = (
                torch.transpose(grad_z_t_1_j, 0, 1)
                .reshape((batch_size * (N - 1), z_dim))
                .detach()
                .clone()
                .to(device)
            )
            E_z_t_1_j = energy[:, t - 1, 1:]
            E_z_t_1_j_flatten = E_z_t_1_j.T.reshape(batch_size * (N - 1))

            Z_t_1_j = Z[:, t - 1, 1:, :]
            # Z_t_1_j_shape = Z_t_1_j.shape
            # p_t_j = (1. - grad_step) * Z_t_1_j - grad_step * beta * grad_z_t_1_j \
            #                                        + eps_scale * v[:, t, 1:, :]
            p_t_j = (
                Z_t_1_j - grad_step * beta * grad_z_t_1_j + eps_scale * v[:, t, 1:, :]
            )

            p_t_j_flatten = (
                torch.transpose(p_t_j, 0, 1)
                .reshape((batch_size * (N - 1), z_dim))
                .detach()
                .clone()
                .to(device)
            )
            p_t_j_flatten.requires_grad_(True)

            Z_t_1_j_flatten = (
                torch.transpose(Z_t_1_j, 0, 1)
                .reshape((batch_size * (N - 1), z_dim))
                .detach()
                .clone()
                .to(device)
            )
            Z_t_1_j_flatten.requires_grad_(True)

            (z_t, E_t, grad_t, mask) = mala_transition.do_transition_step(
                Z_t_1_j_flatten,
                p_t_j_flatten,
                E_z_t_1_j_flatten,
                grad_z_t_1_j_flatten,
                grad_step,
                eps_scale,
                target,
                beta,
            )

            z_t = torch.transpose(z_t.reshape(N - 1, batch_size, z_dim), 0, 1)
            E_t = E_t.reshape(N - 1, batch_size).T
            grad_t = torch.transpose(
                grad_t.reshape(N - 1, batch_size, z_dim),
                0,
                1,
            )
            Z[:, t, 1:, :] = z_t
            energy[:, t, 1:] = E_t.detach().clone()
            grads[:, t, 1:, :] = grad_t.detach().clone()

        log_weights = -(betas_diff[None, :, None] * energy[:, 1:, :]).sum(1)

        max_logs = torch.max(log_weights, dim=1)[0]
        log_weights = log_weights - max_logs[:, None]
        weights = torch.exp(log_weights)
        sum_weights = torch.sum(weights, dim=1)
        weights = weights / sum_weights[:, None]
        weights[weights != weights] = 0.0
        weights[weights.sum(1) == 0.0] = 1.0

        indices = torch.multinomial(weights, 1).squeeze().tolist()
        z = Z[np.arange(batch_size), :, indices, :]
        E = energy[np.arange(batch_size), :, indices]
        grad = grads[np.arange(batch_size), :, indices, :]
        z_sp.append(z[:, -1, :].detach().clone())

    return z_sp, 1.0


def ais_sampling(
    target,
    proposal,
    n_steps,
    grad_step,
    eps_scale,
    n,
    batch_size,
    N,
    betas,
    rhos,
):
    z_last = []
    zs = []
    z = None
    for _ in tqdm(range(0, n, batch_size)):
        z = proposal.sample([batch_size, len(betas)])
        z.requires_grad_(True)

        z_sp, _ = citerais_mala_dynamics(
            z,
            target,
            n_steps,
            grad_step,
            eps_scale,
            N,
            betas,
            rhos,
        )
        last = z_sp[-1].data.cpu().numpy()
        z_last.append(last)
        zs.append(np.stack([o.data.cpu().numpy() for o in z_sp], axis=0))

    z_last_np = np.asarray(z_last).reshape(-1, z.shape[-1])
    zs = np.stack(zs, axis=0)
    return z_last_np, zs


def citerais_ula_dynamics(
    z,
    target,
    proposal,
    n_steps,
    grad_step,
    eps_scale,
    N,
    betas,
    rhos,
    do_ar=True,
    max_n_rej=10,
    pbern=0.5,
):
    """
    do_ar: bool - include path from the privious step to N current paths
    max_n_rej: int - maximum number of rejections together, if reached - don't reject this step (aka refreshing)
    """
    # T = len(betas)
    batch_size, T, z_dim = z.shape
    device = z.device
    z = proposal.sample([batch_size])
    z.requires_grad_(True)
    E, grad = grad_energy(z, target, None)

    # step 0: Generate start ULA path
    start = [z.detach().clone()]
    mala_transition = MALATransition(z_dim, device)
    W = mala_transition.stand_normal.sample([batch_size, T])

    for t in range(1, T):
        beta = betas[::-1][t]
        # rho = rhos[::-1][t]
        z = z - grad_step * beta * grad + eps_scale * W[:, t, :]
        E, grad = grad_energy(z, target, None)
        start.append(z.detach().clone())

    z = torch.stack(start, 1)
    # end of step 0

    z_sp = [z[:, -1, :].clone().detach()]
    # z_sp = [z.clone().detach()]
    traj_hist = [z[:5].clone().detach()]
    # batch_size, T, z_dim = z.shape[0], z.shape[1], z.shape[2]
    T = T - 1  # ??
    # T = len(betas) - 2

    betas = np.array(betas)
    betas_diff = torch.FloatTensor(betas[:-1] - betas[1:]).to(device)  # n-1

    z_flat = (
        torch.transpose(z, 0, 1)
        .reshape((batch_size * z.shape[1], z_dim))
        .detach()
        .clone()
    )
    z_flat.requires_grad_(True)
    E_flat, grad_flat = grad_energy(z_flat, target, None)
    grad = (
        torch.transpose(
            grad_flat.reshape(list(z.shape[:-1][::-1]) + [z.shape[-1]]),
            0,
            1,
        )
        .detach()
        .clone()
    )
    E = E_flat.reshape(z.shape[:-1][::-1]).T.data

    n_rej = torch.zeros(batch_size)
    for _ in trange(n_steps):
        v = torch.zeros((batch_size, T + 1, N, z_dim), dtype=z.dtype).to(
            device,
        )
        # step 1
        kappa = torch.zeros((batch_size, T + 1, z_dim), dtype=z.dtype).to(
            device,
        )
        kappa_t_noise = mala_transition.stand_normal.sample(
            [batch_size, T + 1],
        )
        kappa[:, 0, :] = (
            rhos[-1] * z[:, 0, :]
            + ((1 - rhos[-1] ** 2) ** 0.5) * kappa_t_noise[:, 0, :]
        )

        for t in range(1, T + 1):
            beta = betas[::-1][t]
            rho = rhos[::-1][t]
            v[:, t, 0, :] = (
                z[:, t, :] - z[:, t - 1, :] + grad_step * beta * grad[:, t - 1, :]
            ) / eps_scale
            kappa[:, t, :] = (
                rho * v[:, t, 0, :] + ((1 - rho**2) ** 0.5) * kappa_t_noise[:, t, :]
            )

        # step 2
        kappa_repeat = kappa[:, 0, :].unsqueeze(1).repeat(1, N - 1, 1)
        kappa_N_noise = mala_transition.stand_normal.sample(
            [batch_size, N - 1],
        )
        kappa_repeat_N = kappa.unsqueeze(2).repeat(1, 1, N - 1, 1)

        Z = torch.zeros((batch_size, T + 1, N, z_dim), dtype=z.dtype).to(
            device,
        )
        Z[:, :, 0, :] = z
        mask = (
            torch.distributions.bernoulli.Bernoulli(
                probs=torch.tensor([pbern]),
            )
            .sample(torch.Size([batch_size, N - 1]))
            .squeeze(-1)
            .to(device)
        )

        Z[:, 0, 1:, :] = (
            rhos[-1] * kappa_repeat * mask[..., None]
            + ((1 - rhos[-1] ** 2 * mask[..., None]) ** 0.5) * kappa_N_noise
        )

        kappa_flat = (
            torch.transpose(Z[:, 0, 1:, :], 0, 1)
            .reshape((batch_size * (N - 1), z_dim))
            .detach()
            .clone()
        )
        kappa_E_flat, kappa_grad_flat = grad_energy(kappa_flat, target)
        kappa_E = kappa_E_flat.reshape(N - 1, batch_size).T
        kappa_grad = torch.transpose(
            kappa_grad_flat.reshape(N - 1, batch_size, z_dim),
            0,
            1,
        )

        energy = torch.zeros(batch_size, T + 1, N).to(device)
        energy[:, :, 0] = E.data
        energy[:, 0, 1:] = kappa_E

        grads = torch.zeros(batch_size, T + 1, N, z_dim).to(device)
        grads[:, :, 0, :] = grad.detach().clone()
        grads[:, 0, 1:, :] = kappa_grad

        W_2 = mala_transition.stand_normal.sample([batch_size, T, N - 1])

        for t in range(1, T + 1):
            beta = betas[::-1][t]
            rho = rhos[::-1][t]

            v[:, t, 1:, :] = (
                rho * kappa_repeat_N[:, t, :, :] * mask[..., None]
                + ((1 - rho**2 * mask[..., None]) ** 0.5) * W_2[:, t - 1, :, :]
            )
            # v[:, t, 1:, :] = (rho*kappa_repeat_N[:, t - 1, :, :]*mask[..., None] +
            #                   ((1 - rho**2 * mask[..., None])**0.5) * W_2[:, t - 1, :, :])

            # z_t_1_j_flatten = (torch.transpose(Z[:, t - 1, 1:, :], 0, 1)
            #                    .reshape((batch_size*(N-1), z_dim))
            #                    .detach().clone().to(device))
            # z_t_1_j_flatten.requires_grad_(True)

            grad_z_t_1_j = grads[:, t - 1, 1:, :]

            grad_z_t_1_j_flatten = (
                torch.transpose(grad_z_t_1_j, 0, 1)
                .reshape((batch_size * (N - 1), z_dim))
                .detach()
                .clone()
                .to(device)
            )
            E_z_t_1_j = energy[:, t - 1, 1:]
            E_z_t_1_j_flatten = E_z_t_1_j.T.reshape(batch_size * (N - 1))

            Z_t_1_j = Z[:, t - 1, 1:, :]
            p_t_j = (
                Z_t_1_j - grad_step * beta * grad_z_t_1_j + eps_scale * v[:, t, 1:, :]
            )

            p_t_j_flatten = (
                torch.transpose(p_t_j, 0, 1)
                .reshape((batch_size * (N - 1), z_dim))
                .detach()
                .clone()
                .to(device)
            )
            p_t_j_flatten.requires_grad_(True)

            Z_t_1_j_flatten = (
                torch.transpose(Z_t_1_j, 0, 1)
                .reshape((batch_size * (N - 1), z_dim))
                .detach()
                .clone()
                .to(device)
            )
            Z_t_1_j_flatten.requires_grad_(True)

            (_, E_t, grad_t) = mala_transition.compute_log_probs(
                Z_t_1_j_flatten,
                p_t_j_flatten,
                E_z_t_1_j_flatten,
                grad_z_t_1_j_flatten,
                grad_step,
                eps_scale,
                target,
                beta,
            )
            z_t = p_t_j_flatten
            z_t = torch.transpose(z_t.reshape(N - 1, batch_size, z_dim), 0, 1)
            E_t = E_t.reshape(N - 1, batch_size).T
            grad_t = torch.transpose(
                grad_t.reshape(N - 1, batch_size, z_dim),
                0,
                1,
            )
            Z[:, t, 1:, :] = z_t
            energy[:, t, 1:] = E_t.detach().clone()
            grads[:, t, 1:, :] = grad_t.detach().clone()

        traj_hist.append(Z[:5, ..., 1, :].detach().clone())

        log_weights = -(betas_diff[None, :, None] * energy[:, 1:, :]).sum(1)

        max_logs = torch.max(log_weights, dim=1)[0]
        log_weights = log_weights - max_logs[:, None]
        weights = torch.exp(log_weights)
        sum_weights = torch.sum(weights, dim=1)
        weights = weights / sum_weights[:, None]
        weights[weights != weights] = 0.0
        weights[weights.sum(1) == 0.0] = 1.0

        if not do_ar:
            weights[:, 0, ...] = 0.0

        indices = torch.multinomial(weights, 1).squeeze().tolist()
        n_rej += (torch.tensor(indices) == 0).float()

        weights[n_rej > max_n_rej, 0, ...] = 0.0
        indices = torch.multinomial(weights, 1).squeeze().tolist()
        n_rej[n_rej > max_n_rej] = 0

        z = Z[np.arange(batch_size), :, indices, :]
        E = energy[np.arange(batch_size), :, indices]
        grad = grads[np.arange(batch_size), :, indices, :]
        z_sp.append(z[:, -1, :].detach().clone())
        # z_sp.append(z.detach().clone())

    return z_sp, 1.0, traj_hist


def citerais_ula_sampling(
    target,
    proposal,
    n_steps,
    grad_step,
    eps_scale,
    n,
    batch_size,
    N,
    betas,
    rhos,
):
    z_last = []
    zs = []
    z = None
    for _ in tqdm(range(0, n, batch_size)):
        z = proposal.sample([batch_size, len(betas)])
        z.requires_grad_(True)
        z_sp, _, _ = citerais_ula_dynamics(
            z,
            target,
            proposal,
            n_steps,
            grad_step,
            eps_scale,
            N,
            betas,
            rhos,
        )
        last = z_sp[-1].data.cpu().numpy()
        z_last.append(last)
        zs.append(np.stack([o.data.cpu().numpy() for o in z_sp], axis=0))

    z_last_np = np.asarray(z_last).reshape(-1, z.shape[-1])
    zs = np.stack(zs, axis=0)
    return z_last_np, zs



import copy
from typing import Dict, Iterable, List, Optional, Tuple, Union

import torch
from torch import nn
from torch.distributions import Distribution as torchDist
from tqdm import trange

from .gan_distribution import Distribution
from .samplers import MCMCRegistry
from .utils import time_comp_cls
from .utils.callbacks import Callback


class Sampler:
    def __init__(
        self,
        gen: nn.Module,
        ref_dist: Union[Distribution, torchDist],
        *,
        burn_in_steps: int = 0,
        start_sample: int = 0,
        n_sampling_steps: int = 1,
        batch_size: Optional[int] = None,
        save_every: int = 1,
        verbose: bool = True,
        n_steps: Optional[int] = None,
        collect_imgs: bool = True,
        sampling: str = "ula",
        mcmc_args: Optional[Dict] = None,
        callbacks: Optional[Iterable[Callback]] = None,
        keep_graph: bool = False,
    ):
        self.gen = gen
        self._ref_dist = ref_dist
        self.burn_in_steps = burn_in_steps
        self.start_sample = start_sample
        self.n_sampling_steps = n_sampling_steps
        self.save_every = save_every
        self.verbose = verbose
        self.trange = trange if self.verbose else range
        self.n_steps = n_steps
        self.collect_imgs = collect_imgs
        self.callbacks = callbacks or []
        self.keep_graph = keep_graph

        self.sampling = sampling
        self.init_mcmc_args: Dict = copy.deepcopy(mcmc_args or dict())
        self.mcmc_args = copy.deepcopy(self.init_mcmc_args)

        self.mcmc = MCMCRegistry()
        self.target = self.ref_dist

    def reset(self):
        self.mcmc_args = copy.deepcopy(self.init_mcmc_args)
        for callback in self.callbacks:
            callback.reset()

    @property
    def ref_dist(self) -> Distribution:
        return self._ref_dist

    def step(
        self,
        z: torch.Tensor,
        it: int = 1,
        # data_batch: Optional[torch.FloatTensor] = None,
        meta: Optional[Dict] = None,
        keep_graph: bool = False,
    ) -> Tuple[torch.Tensor, Dict]:
        # self._ref_dist.data_batch = data_batch
        pts, meta = self.mcmc(
            self.sampling,
            z,
            self.target,
            proposal=self.gen.proposal,
            # n_samples=self.n_sampling_steps,
            # burn_in=self.n_sampling_steps - 1,
            project=self.ref_dist.project,
            **self.mcmc_args,
            meta=meta,
            keep_graph=keep_graph,
        )

        self.mcmc_args.update(
            {key: meta[key][-1] for key in self.mcmc_args.keys() & meta.keys()}
        )

        return pts[-1], meta

    @time_comp_cls
    def __call__(
        self,
        z: torch.Tensor,
        n_steps: Optional[int] = None,
        # data_batch: Optional[torch.FloatTensor] = None,
        collect_imgs: bool = False,
        keep_graph: bool = False,
    ) -> Tuple[List, List, List, List]:
        n_steps = n_steps if n_steps is not None else self.n_steps
        collect_imgs = collect_imgs or self.collect_imgs
        keep_graph = keep_graph or self.keep_graph
        zs = [z.cpu()]
        xs = []
        meta = dict()
        if collect_imgs:
            xs.append(self.gen.inverse_transform(self.gen(z)).detach().cpu())

        it = 0
        for it in self.trange(1, n_steps + 1):
            new_z, meta = self.step(z, it, meta=meta, keep_graph=keep_graph)
            if it > self.start_sample:
                z = new_z

            if it > self.burn_in_steps and it % self.save_every == 0:
                if keep_graph:
                    zs.append(z.cpu())
                else:
                    zs.append(z.detach().cpu())
                if collect_imgs:
                    xs.append(
                        self.gen.inverse_transform(self.gen(z.detach())).detach().cpu()
                    )

            for callback in self.callbacks:
                callback.invoke(self.mcmc_args)

        return zs, xs



import copy
from dataclasses import dataclass
from pathlib import Path

import numpy as np
import pandas as pd
import torch
from sklearn import datasets, preprocessing
from sklearn.ensemble import IsolationForest
from sklearn.model_selection import train_test_split


def filter_outliers(X, random_state=42):
    clf = IsolationForest(random_state=random_state)
    # clf = LocalOutlierFactor(n_neighbors=20) #, contamination=0.1)
    # use fit_predict to compute the predicted labels of the training samples
    # (when LOF is used for outlier detection, the estimator has no predict,
    # decision_function and score_samples methods).
    y_pred = clf.fit_predict(X)
    X = X[y_pred == 1]

    return y_pred


def standartize(X_train, X_test, intercept=True):
    """Whitens noise structure, covariates updated"""
    X_train = copy.deepcopy(X_train)
    X_test = copy.deepcopy(X_test)
    # if intercept: #adds intercept term
    #     X_train = np.concatenate((np.ones(X_train.shape[0]).reshape(X_train.shape[0],1),X_train),axis=1)
    #     X_test = np.concatenate((np.ones(X_test.shape[0]).reshape(X_test.shape[0],1),X_test),axis=1)
    # d = X_train.shape[1]

    # Centering the covariates
    means = np.mean(X_train, axis=0)
    if intercept:  # do not subtract the mean from the bias term
        means[0] = 0.0
    # Normalizing the covariates
    X_train -= means
    Cov_matr = np.dot(X_train.T, X_train)
    U, S, V_T = np.linalg.svd(Cov_matr, compute_uv=True)
    # Sigma_half = U @ np.diag(np.sqrt(S)) @ V_T
    Sigma_minus_half = U @ np.diag(1.0 / np.sqrt(S)) @ V_T
    X_train = X_train @ Sigma_minus_half
    # The same for test sample
    X_test = (X_test - means) @ Sigma_minus_half
    return X_train, X_test


def split(X, y, c1, c2):
    X = X.astype("float32")
    y = y.astype("int")

    # Select the classes to classify for binary logistic regression
    i1 = np.where((y == c1) | (y == c2))
    y = y[i1]
    X = X[i1]

    # Replace labels of the current class with 1 and other labels with 0
    i1 = np.where(y == c1)
    i2 = np.where(y == c2)
    y[i1] = 1
    y[i2] = 0

    x_train, x_test, y_train, y_test = train_test_split(
        X,
        y,
        test_size=0.2,
        random_state=42,
    )

    return x_train, x_test, y_train, y_test


def preprocess(x_train, x_test, y_train, y_test, normalize=False):
    # Making sure that the values are float so that we can get decimal points
    # after division

    outliers_pred = filter_outliers(x_train)
    print(
        f"Original size: {len(x_train)}, outliers size: {sum(outliers_pred == -1)}",
    )

    x_train = x_train[outliers_pred == 1]
    y_train = y_train[outliers_pred == 1]

    min_max_scaler = preprocessing.MinMaxScaler()

    x_train = min_max_scaler.fit_transform(x_train)
    x_test = min_max_scaler.transform(x_test)

    x_train = np.concatenate(
        (np.ones(x_train.shape[0]).reshape(x_train.shape[0], 1), x_train),
        axis=1,
    )
    x_test = np.concatenate(
        (np.ones(x_test.shape[0]).reshape(x_test.shape[0], 1), x_test),
        axis=1,
    )

    if normalize:
        x_train, x_test = standartize(x_train, x_test, intercept=True)

    # Return objects of interest
    return x_train, x_test, y_train, y_test


def import_covertype(c1=1, c2=2, n_data=500000):
    dataset = datasets.fetch_covtype()
    # min_max_scaler = preprocessing.MinMaxScaler()
    # X = min_max_scaler.fit_transform(dataset.data)
    X = dataset.data
    y = dataset.target
    # x_train = X[0:n_data,:]
    # x_test = X[n_data:,:]
    # y_train = dataset.target[0:n_data]
    # y_test = dataset.target[n_data:]
    x_train, x_test, y_train, y_test = split(X, y, c1, c2)

    return preprocess(x_train, x_test, y_train, y_test, normalize=False)


def import_breast(c1=0, c2=1, n_data=569):
    dataset = datasets.load_breast_cancer()
    # min_max_scaler = preprocessing.MinMaxScaler()
    # X = min_max_scaler.fit_transform(dataset.data)
    X = dataset.data
    y = dataset.target

    # X = dataset.data
    # x_train = X[0:n_data,:]
    # x_test = X[n_data:,:]
    # y_train = dataset.target[0:n_data]
    # y_test = dataset.target[n_data:]
    x_train, x_test, y_train, y_test = split(X, y, c1, c2)

    return preprocess(x_train, x_test, y_train, y_test)


def import_csv_dataset(path, c1=0, c2=1):
    dataset = pd.read_csv(path, delimiter=",")
    dataset = dataset.to_numpy()
    X = dataset[:, :-1]
    y = dataset[:, -1]
    x_train, x_test, y_train, y_test = split(X, y, c1, c2)

    return preprocess(x_train, x_test, y_train, y_test)


def import_digits(c1=5, c2=6):
    dataset = datasets.load_digits()
    # min_max_scaler = preprocessing.MinMaxScaler()
    # X = min_max_scaler.fit_transform(dataset.data)
    X = dataset.data
    y = dataset.target

    # X = dataset.data
    # x_train = X[0:n_data,:]
    # x_test = X[n_data:,:]
    # y_train = dataset.target[0:n_data]
    # y_test = dataset.target[n_data:]
    x_train, x_test, y_train, y_test = split(X, y, c1, c2)

    return preprocess(x_train, x_test, y_train, y_test)


@dataclass
class ClassificationDataset:
    x_train: torch.Tensor
    x_test: torch.Tensor
    y_train: torch.Tensor
    y_test: torch.Tensor

    @property
    def d(self):
        return self.x_train.shape[1]

    @property
    def n(self):
        self.y_train.shape[0]


class ClassificationDatasetFactory:
    def __init__(self, data_root=None, device="cpu", **kwargs):
        self._instance = None
        self.data_root = data_root
        self.device = device

    def get_dataset(self, name: str, **kwargs) -> ClassificationDataset:
        if name == "covertype":
            x_train, x_test, y_train, y_test = import_covertype(**kwargs)
        elif name == "breast":
            x_train, x_test, y_train, y_test = import_breast(**kwargs)
        elif name == "digits":
            x_train, x_test, y_train, y_test = import_digits(**kwargs)
        else:
            # csv
            if not name.endswith(".csv"):
                name = f"{name}.csv"
            path = Path(self.data_root, name)
            x_train, x_test, y_train, y_test = import_csv_dataset(
                path,
                **kwargs,
            )

        x_train = torch.tensor(
            x_train,
            dtype=torch.float32,
            device=self.device,
        )
        x_test = torch.tensor(x_test, dtype=torch.float32, device=self.device)
        y_train = torch.tensor(
            y_train,
            dtype=torch.float32,
            device=self.device,
        )
        y_test = torch.tensor(y_test, dtype=torch.float32, device=self.device)

        return ClassificationDataset(x_train, x_test, y_train, y_test)


# Import data

"""
w, v = np.linalg.eig(np.dot(X.T,X))
w = np.max(w.real)
M = w * (1/4) + tau
gamma = 0.01/M

# Initialization
d = np.size(X,1)
n = np.size(y)
tau = 1 # regularization parameter
"""



import random
import time
from collections.abc import Mapping
from pathlib import Path

import numpy as np
import torch


PROJECT_PATH = Path(__file__).parent.parent.parent
CHECKPOINT_DIR = Path(PROJECT_PATH, "checkpoints")
LOG_DIR = Path(PROJECT_PATH, "log")
IMGS_DIR = Path(PROJECT_PATH, "imgs")


def random_seed(seed):
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    np.random.seed(seed)
    random.seed(seed)
    g = torch.Generator()
    g.manual_seed(seed)


def seed_worker(worker_id):
    worker_seed = torch.initial_seed() % 2**32
    np.random.seed(worker_seed)
    random.seed(worker_seed)


class IgnoreLabelDataset(torch.utils.data.Dataset):
    def __init__(self, orig):
        self.orig = orig

    def __getitem__(self, index):
        return self.orig[index][0]

    def __len__(self):
        return len(self.orig)


def time_comp(fun):
    def wrapper(*args, **kwargs):
        t0 = time.perf_counter()
        result = fun(*args, **kwargs)
        tf = time.perf_counter()
        dt = tf - t0
        if kwargs.get("verbose"):
            print("Time elapsed: ", "{:.2f}".format(dt) + "s")

        return result

    return wrapper


def time_comp_cls(fun):
    def wrapper(*args, **kwargs):
        t0 = time.perf_counter()
        result = fun(*args, **kwargs)
        tf = time.perf_counter()
        dt = tf - t0
        if getattr(args[0], "verbose", False):
            print("Time elapsed: ", "{:.2f}".format(dt) + "s")

        return result

    return wrapper


class DotConfig(Mapping):
    """
    Simple wrapper for config
    allowing access with dot notation
    """

    def __init__(self, yaml):
        self._dict = yaml

    def __getattr__(self, key):
        if key in self.__dict__:
            return super().__getattr__(key)
        if key in self._dict:
            value = self._dict[key]
            if isinstance(value, dict):
                return DotConfig(value)
            return value
        else:
            return None

    def items(self):
        return [(k, DotConfig(v)) for k, v in self._dict.items()]

    def keys(self):
        return self._dict.keys()

    def __len__(self):
        return len(self._dict)

    def __iter__(self):
        return self._dict.__iter__()

    def __getitem__(self, key):
        return self._dict[key]

    @property
    def dict(self):
        return self._dict

    def __contains__(self, key):
        return key in self._dict

    def __setitem__(self, key, value):
        self._dict[key] = value

    # def __assign__(self, key, value):
    #     print('ho')
    #     self._dict[key] = value

    # def __setattr__(self, key, value):
    #     print(key)
    #     self._dict[key] = value



# Packages
from dataclasses import dataclass
from pathlib import Path

import torch

from .logistic_regression import import_csv_dataset


@dataclass
class RegressionDataset:
    x_train: torch.Tensor
    x_test: torch.Tensor
    y_train: torch.Tensor
    y_test: torch.Tensor

    @property
    def d(self):
        return self.x_train.shape[1]

    @property
    def n(self):
        self.y_train.shape[0]


class RegressionDatasetFactory:
    def __init__(self, data_root=None, device="cpu", **kwargs):
        self._instance = None
        self.data_root = data_root
        self.device = device

    def get_dataset(self, name: str, **kwargs) -> RegressionDataset:
        if name == "some_dataset_name":
            pass
        else:
            # csv
            if not name.endswith(".csv"):
                name = f"{name}.csv"
            path = Path(self.data_root, name)
            x_train, x_test, y_train, y_test = import_csv_dataset(
                path,
                **kwargs,
            )

        x_train = torch.tensor(
            x_train,
            dtype=torch.float32,
            device=self.device,
        )
        x_test = torch.tensor(x_test, dtype=torch.float32, device=self.device)
        y_train = torch.tensor(
            y_train,
            dtype=torch.float32,
            device=self.device,
        )
        y_test = torch.tensor(y_test, dtype=torch.float32, device=self.device)

        return RegressionDataset(x_train, x_test, y_train, y_test)



from abc import ABC, abstractmethod

import torch


def increment_steps(call):
    def wrapper(self, start: torch.Tensor, target, proposal, *args, **kwargs):
        if "n_steps" in kwargs:
            n_steps = kwargs.get("n_steps")
        elif len(args) > 0:
            n_steps = args[0]
        elif "n_steps" in self.__dict__:
            n_steps = self.__dict__["n_steps"]
        else:
            raise TypeError("Missing argument 'n_steps'")
        out = call(self, start, target, proposal, n_steps, *args, **kwargs)
        self._steps_done += n_steps
        return out

    return wrapper


def adapt_stepsize_dec(call):
    def wrapper(self, *args, **kwargs):
        adapt_stepsize = kwargs.get("adapt_stepsize")

        if (
            self.adapt_stepsize is True and adapt_stepsize is not False
        ) or adapt_stepsize is True:
            adapt_stepsize = True
        else:
            adapt_stepsize = False

        if adapt_stepsize and self._steps_done > 0:
            grad_step = self.grad_step
            noise_scale = (2 * self.grad_step) ** 0.5

            kwargs["grad_step"] = grad_step
            kwargs["noise_scale"] = noise_scale

        out = call(self, *args, **kwargs)
        grad_step = out[-1]

        if adapt_stepsize:
            self.grad_step = grad_step

        return out

    return wrapper


class AbstractMCMC(ABC):
    _steps_done = 0

    def __init__(self, verbose=True, **kwargs):
        self._steps_done = 0
        self.verbose = verbose

    @abstractmethod
    @increment_steps
    @adapt_stepsize_dec
    def __call__(
        self,
        start: torch.Tensor,
        target,
        proposal,
        n_steps: int,
        *args,
        **kwargs,
    ):
        raise NotImplementedError

    @property
    def steps_done(self):
        return self._steps_done



import torch
from torch import nn


def get_optimizer(parameters, optimizer="Adam", lr=1e-3, weight_decay=1e-5):
    if optimizer == "Adam":
        return torch.optim.Adam(parameters, lr=lr, weight_decay=weight_decay)
    else:
        raise NotImplementedError


def get_loss(loss):
    if loss == "mix_kl":
        return MixKLLoss
    if loss == "forward_kl":
        return forward_kl
    if loss == "backward_kl":
        return backward_kl
    else:
        raise NotImplementedError


def entropy(target, proposal, flow, y, acc_rate=1.0):
    u = proposal.sample(y.shape[:-1])
    x, log_jac = flow.inverse(u)
    entr = -proposal.log_prob(u) + log_jac
    entr = acc_rate * entr
    grad_est = entr

    return entr.mean(), grad_est.mean()


def forward_kl(target, proposal, flow, x):
    x_ = x.detach().requires_grad_()
    z, log_jac = flow.forward(x_)
    est = target.log_prob(x_) - (proposal.log_prob(z) + log_jac)
    grad_est = -proposal.log_prob(z) - log_jac
    return est.mean(), grad_est.mean()


def backward_kl(target, proposal, flow, x):
    z = proposal.sample(x.shape[:-1])
    x, log_jac_inv = flow.inverse(z)
    est = proposal.log_prob(z) - log_jac_inv - target.log_prob(x)
    grad_est = -log_jac_inv - target.log_prob(x)
    return est.mean(), grad_est.mean()


def mix_kl(
    target,
    proposal,
    flow,
    y,
    alpha=0.99,
    gamma=None,
):
    est_f, grad_est_f = forward_kl(target, proposal, flow, y)
    est_b, grad_est_b = backward_kl(target, proposal, flow, y)

    if torch.isnan(grad_est_b).item():
        grad_est_b = 0

    return (
        alpha * est_f + (1.0 - alpha) * est_b,
        alpha * grad_est_f + (1.0 - alpha) * grad_est_b,
    )


class MixKLLoss(nn.Module):
    def __init__(
        self,
        target,
        proposal,
        flow,
        alpha=0.99,
        beta=0.0,
        gamma=0.99,
    ):  # .2):#.99):
        super().__init__()
        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma
        self.flow = flow
        self.target = target
        self.proposal = proposal

    def forward(self, y, acc_rate=1.0, alpha=None, beta=None, gamma=None):
        alpha = alpha if alpha is not None else self.alpha
        beta = beta if beta is not None else self.beta
        gamma = gamma if gamma is not None else self.gamma

        return mix_kl(
            self.target,
            self.proposal,
            self.flow,
            y,
            alpha=alpha,
        )



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Callable, Dict, List, Optional, Union

import numpy as np
import torch
from matplotlib import pyplot as plt
from torchvision import transforms
from torchvision.utils import make_grid


class Callback(ABC):
    cnt: int = 0

    @abstractmethod
    def invoke(self, info: Dict[str, Union[float, np.ndarray]]):
        raise NotImplementedError

    def reset(self):
        self.cnt = 0


class CallbackRegistry:
    registry = {}

    @classmethod
    def register(cls, name: Optional[str] = None) -> Callable:
        def inner_wrapper(wrapped_class: Callback) -> Callback:
            if name is None:
                name_ = wrapped_class.__name__
            else:
                name_ = name
            cls.registry[name_] = wrapped_class
            return wrapped_class

        return inner_wrapper

    @classmethod
    def create(cls, name: str, **kwargs) -> Callback:
        model = cls.registry[name]
        model = model(**kwargs)
        return model


@CallbackRegistry.register()
class WandbCallback(Callback):
    def __init__(
        self,
        *,
        invoke_every: int = 1,
        init_params: Optional[Dict] = None,
        keys: Optional[List[str]] = None,
    ):
        self.init_params = init_params if init_params else {}
        import wandb

        self.wandb = wandb
        wandb.init(**self.init_params)

        self.invoke_every = invoke_every
        self.keys = keys

        self.img_transform = transforms.Resize(
            128, interpolation=transforms.InterpolationMode.NEAREST
        )

    def invoke(self, info: Dict[str, Union[float, np.ndarray]]):
        step = info.get("step", self.cnt)
        if step % self.invoke_every == 0:
            wandb = self.wandb
            if not self.keys:
                self.keys = info.keys()
            log = dict()
            for key in self.keys:
                if key not in info:
                    continue
                if isinstance(info[key], np.ndarray):
                    log[key] = wandb.Image(
                        make_grid(
                            self.img_transform(
                                torch.clip(torch.from_numpy(info[key][:25]), 0, 1)
                            ),
                            nrow=5,
                        ),
                        caption=key,
                    )
                else:
                    log[key] = info[key]
            log["step"] = step
            wandb.log(log)
        self.cnt += 1
        return 1

    def reset(self):
        super().reset()
        self.wandb.init(**self.init_params)


@CallbackRegistry.register()
class PlotImagesCallback(Callback):
    def __init__(
        self,
        *,
        invoke_every: int = 1,
    ):
        self.invoke_every = invoke_every
        self.img_transform = transforms.Resize(
            128, interpolation=transforms.InterpolationMode.NEAREST
        )

    def invoke(self, info: Dict[str, Union[float, np.ndarray]]):
        step = info.get("step", self.cnt)
        if step % self.invoke_every == 0:
            grid = make_grid(
                self.img_transform(
                    torch.clip(torch.from_numpy(info["imgs"][:100]), 0, 1)
                ),
                nrow=10,
            ).permute(1, 2, 0)
            info["grid"] = grid.numpy()
        self.cnt += 1

        return 1


@CallbackRegistry.register()
class SaveImagesCallback(Callback):
    def __init__(
        self,
        save_dir: Union[Path, str],
        *,
        invoke_every: int = 1,
    ):
        self.invoke_every = invoke_every
        self.save_dir = save_dir

    def invoke(self, info: Dict[str, Union[float, np.ndarray]]):
        if self.cnt % self.invoke_every == 0:
            imgs = info["imgs"]
            savepath = Path(self.save_dir, f"imgs_{self.cnt}.npy")
            np.save(savepath, imgs)

        self.cnt += 1


@CallbackRegistry.register()
class DiscriminatorCallback(Callback):
    def __init__(
        self,
        gan,
        *,
        invoke_every=1,
        update_input=True,
        device="cuda",
        batch_size: Optional[int] = None,
    ):
        self.invoke_every = invoke_every
        self.dis = gan.dis
        self.transform = self.dis.transform
        self.update_input = update_input
        self.device = device
        self.batch_size = batch_size

    @torch.no_grad()
    def invoke(
        self,
        info: Dict[str, Union[float, np.ndarray]],
        batch_size: Optional[int] = None,
    ):
        dgz = None
        if self.cnt % self.invoke_every == 0:
            imgs = info["imgs"]
            if "label" in info:
                label = torch.LongTensor(info["label"]).to(self.device)
            else:
                label = None

            if not batch_size:
                batch_size = len(imgs) if not self.batch_size else self.batch_size
            x = self.transform(torch.from_numpy(imgs).to(self.device))
            dgz = 0
            for i, x_batch in enumerate(torch.split(x, batch_size)):
                if label is not None:
                    label_batch = label[i * batch_size : (i + 1) * batch_size]
                else:
                    label_batch = None
                self.dis.label = label_batch
                # dgz += self.dis.output_layer(self.dis(x_batch)).sum().item()
                dgz += (self.dis(x_batch).squeeze()).sum().item()
            dgz /= len(imgs)

            if self.update_input:
                info["D(G(z))"] = dgz
        self.cnt += 1

        return dgz


@CallbackRegistry.register()
class EnergyCallback(Callback):
    def __init__(
        self,
        gan,
        *,
        invoke_every=1,
        update_input=True,
        device="cuda",
        norm_constant=1,
        batch_size: Optional[int] = None,
        log_norm_const: float = 0,
    ):
        self.invoke_every = invoke_every
        self.dis = gan.dis
        self.gen = gan.gen
        self.norm_constant = norm_constant
        self.transform = self.dis.transform
        self.update_input = update_input
        self.device = device
        self.batch_size = batch_size
        self.log_norm_const = log_norm_const

    @torch.no_grad()
    def invoke(
        self,
        info: Dict[str, Union[float, np.ndarray]],
        batch_size: Optional[int] = None,
    ):
        energy = None
        if self.cnt % self.invoke_every == 0:
            zs = torch.FloatTensor(info["zs"]).to(self.device)
            if "label" in info:
                label = torch.LongTensor(info["label"]).to(self.device)
            else:
                label = None

            if not batch_size:
                batch_size = len(zs) if not self.batch_size else self.batch_size
            energy = 0

            for i, z_batch in enumerate(torch.split(zs, batch_size)):
                if label is not None:
                    label_batch = label[i * batch_size : (i + 1) * batch_size]
                    self.dis.label = label_batch
                    self.gen.label = label_batch
                else:
                    label_batch = None

                dgz = self.dis(self.gen(z_batch)).squeeze()
                energy += -(self.gen.prior.log_prob(z_batch).sum() + dgz.sum()).item()
            energy /= len(zs)
            energy += self.log_norm_const

            if self.update_input:
                info["Energy"] = energy
        self.cnt += 1

        return energy


@CallbackRegistry.register()
class LogCallback(Callback):
    def __init__(
        self,
        save_dir: Union[Path, str],
        keys: List[str],
        *,
        invoke_every: int = 1,
        resume=False,
    ):
        self.save_dir = Path(save_dir)
        self.invoke_every = invoke_every
        self.keys = keys

        self.save_paths = []
        for key in keys:
            path = Path(save_dir, f"{key}.txt")
            if not resume:
                path.open("w")
            self.save_paths.append(path)

    @torch.no_grad()
    def invoke(
        self,
        info: Dict[str, Union[float, np.ndarray]],
    ):
        step = info.get("step", self.cnt)
        if step % self.invoke_every == 0:
            for save_path, key in zip(self.save_paths, self.keys):
                if key not in info:
                    continue
                if isinstance(info[key], np.ndarray):
                    save_dir = Path(self.save_dir, key)
                    save_dir.mkdir(exist_ok=True)
                    step = info.get("step", 0)
                    save_path = Path(save_dir, f"{key}_{step}.png")
                    self.plot(info[key], save_path)
                    save_path = Path(save_dir, f"{key}_{step}.pdf")
                    self.plot(info[key], save_path)
                else:
                    with save_path.open("ab") as f:
                        np.savetxt(f, [info[key]], delimiter=" ", newline=" ")
        self.cnt += 1
        return 1

    def reset(self):
        super().reset()
        for save_path in self.save_paths:
            save_path.open("ab").write(b"\n")

    @staticmethod
    def plot(imgs, save_path):
        if not isinstance(imgs, list):
            imgs = [imgs]
        fig, axs = plt.subplots(ncols=len(imgs), squeeze=False)
        for i, img in enumerate(imgs):
            axs[0, i].imshow(img)
            axs[0, i].set(xticklabels=[], yticklabels=[], xticks=[], yticks=[])
            axs[0, i].axis("off")
        fig.tight_layout()
        plt.savefig(save_path, bbox_inches="tight")


@CallbackRegistry.register()
class Plot2dEnergyCallback(Callback):
    def __init__(
        self,
        gan,
        save_dir: Union[str, Path],
        # every: int,
        *,
        invoke_every: int = 1,
        device="cuda",
    ) -> None:
        self.save_dir = Path(save_dir, "figs")
        self.save_dir.mkdir(exist_ok=True)
        # self.modes = modes
        self.invoke_every = invoke_every
        # self.every = every
        self.device = device
        self.gan = gan

    def invoke(self, info: Dict[str, Union[float, np.ndarray]]):
        step = info.get("step", self.cnt)
        if step % self.invoke_every == 0:
            n_pts_ax = 100
            real_grid = np.meshgrid(
                np.linspace(-5, 5, n_pts_ax), np.linspace(-5, 5, n_pts_ax)
            )
            reals = (
                torch.from_numpy(np.stack(real_grid, -1).reshape(-1, 2))
                .to(self.device)
                .float()
            )
            dgz = (
                self.gan.dis(self.gan.transform(reals)).squeeze().detach().cpu().numpy()
            )
            xs_grid = reals.detach().cpu().numpy().reshape(n_pts_ax, n_pts_ax, 2)

            fig = plt.figure()
            ax = fig.add_subplot(111)
            im = ax.contourf(
                xs_grid[..., 0],
                xs_grid[..., 1],
                dgz.reshape(n_pts_ax, n_pts_ax),
                levels=100,
            )
            fig.colorbar(im)
            plt.axis("equal")
            plt.axis("off")

            savepath = Path(
                self.save_dir,
                f"{self.save_dir.parts[-2]}_2d_{step}_dgz",
            ).as_posix()
            plt.savefig(savepath + ".png")
            plt.savefig(savepath + ".pdf")
            plt.close()

        self.cnt += 1

        return 1



from typing import Iterable, Optional, Union

import torch
import torch.nn as nn
from pyro.distributions.transforms import AffineCoupling
from torch.distributions import MultivariateNormal as MNormal


class ConditionalDenseNN(nn.Module):
    """
    An implementation of a simple dense feedforward network taking a context variable, for use
      in, e.g.,
    some conditional flows such as :class:`pyro.distributions.transforms.ConditionalAffineCoupling`.

    Example usage:

    >>> input_dim = 10
    >>> context_dim = 5
    >>> x = torch.rand(100, input_dim)
    >>> z = torch.rand(100, context_dim)
    >>> nn = ConditionalDenseNN(input_dim, context_dim, [50], param_dims=[1, input_dim, input_dim])
    >>> a, b, c = nn(x, context=z)  # parameters of size (100, 1), (100, 10), (100, 10)

    :param input_dim: the dimensionality of the input
    :type input_dim: int
    :param context_dim: the dimensionality of the context variable
    :type context_dim: int
    :param hidden_dims: the dimensionality of the hidden units per layer
    :type hidden_dims: list[int]
    :param param_dims: shape the output into parameters of dimension (p_n,) for p_n in param_dims
        when p_n > 1 and dimension () when p_n == 1. The default is [1, 1], i.e. output two parameters of dimension ().
    :type param_dims: list[int]
    :param nonlinearity: The nonlinearity to use in the feedforward network such as torch.nn.ReLU(). Note that no
        nonlinearity is applied to the final network output, so the output is an unbounded real number.
    :type nonlinearity: torch.nn.Module

    """

    def __init__(
        self,
        input_dim,
        context_dim,
        hidden_dims,
        param_dims=[1, 1],
        nonlinearity=torch.nn.ReLU(),
        init_weight_scale=1e-3,
    ):
        super().__init__()

        self.input_dim = input_dim
        self.context_dim = context_dim
        self.hidden_dims = hidden_dims
        self.param_dims = param_dims
        self.count_params = len(param_dims)
        self.output_multiplier = sum(param_dims)
        self.init_weight_scale = init_weight_scale

        # Calculate the indices on the output corresponding to each parameter
        ends = torch.cumsum(torch.tensor(param_dims), dim=0)
        starts = torch.cat((torch.zeros(1).type_as(ends), ends[:-1]))
        self.param_slices = [slice(s.item(), e.item()) for s, e in zip(starts, ends)]

        # Create masked layers
        layers = [torch.nn.Linear(input_dim + context_dim, hidden_dims[0])]
        self.init_params(layers[0].weight)

        for i in range(1, len(hidden_dims)):
            layers.append(torch.nn.Linear(hidden_dims[i - 1], hidden_dims[i]))
            self.init_params(layers[-1].weight)
        layers.append(torch.nn.Linear(hidden_dims[-1], self.output_multiplier))
        self.init_params(layers[-1].weight)
        self.layers = torch.nn.ModuleList(layers)

        # Save the nonlinearity
        self.f = nonlinearity

    def forward(self, x, context):
        # We must be able to broadcast the size of the context over the input
        context = context.expand(x.size()[:-1] + (context.size(-1),))

        x = torch.cat([context, x], dim=-1)
        return self._forward(x)

    def init_params(self, params):
        # torch.nn.init.xavier_uniform_(params, gain=nn.init.calculate_gain('relu')) #45
        torch.nn.init.sparse_(params, sparsity=0.3, std=self.init_weight_scale)
        # torch.nn.init.normal_(params, 0, self.init_weight_scale) 46

    def _forward(self, x):
        """
        The forward method
        """
        h = x
        for layer in self.layers[:-1]:
            h = self.f(layer(h))
        h = self.layers[-1](h)

        # Shape the output, squeezing the parameter dimension if all ones
        if self.output_multiplier == 1:
            return h
        else:
            h = h.reshape(list(x.size()[:-1]) + [self.output_multiplier])

            if self.count_params == 1:
                return h

            else:
                return tuple(h[..., s] for s in self.param_slices)


class DenseNN(ConditionalDenseNN):
    """
    An implementation of a simple dense feedforward network, for use in, e.g., some conditional
      flows such as
    :class:`pyro.distributions.transforms.ConditionalPlanarFlow` and other unconditional flows
      such as
    :class:`pyro.distributions.transforms.AffineCoupling` that do not require an autoregressive
      network.

    Example usage:

    >>> input_dim = 10
    >>> context_dim = 5
    >>> z = torch.rand(100, context_dim)
    >>> nn = DenseNN(context_dim, [50], param_dims=[1, input_dim, input_dim])
    >>> a, b, c = nn(z)  # parameters of size (100, 1), (100, 10), (100, 10)

    :param input_dim: the dimensionality of the input
    :type input_dim: int
    :param hidden_dims: the dimensionality of the hidden units per layer
    :type hidden_dims: list[int]
    :param param_dims: shape the output into parameters of dimension (p_n,) for p_n in param_dims
        when p_n > 1 and dimension () when p_n == 1. The default is [1, 1], i.e. output two
      parameters of dimension ().
    :type param_dims: list[int]
    :param nonlinearity: The nonlinearity to use in the feedforward network such as
      torch.nn.ReLU(). Note that no
        nonlinearity is applied to the final network output, so the output is an unbounded
        real number.
    :type nonlinearity: torch.nn.module

    """

    def __init__(
        self,
        input_dim,
        hidden_dims,
        param_dims=[1, 1],
        nonlinearity=torch.nn.ReLU(),
        init_weight_scale=1e-3,
    ):
        super().__init__(
            input_dim,
            0,
            hidden_dims,
            param_dims=param_dims,
            nonlinearity=nonlinearity,
            init_weight_scale=init_weight_scale,
        )

    def forward(self, x):
        return self._forward(x)


class RNVP(nn.Module):
    def __init__(
        self,
        num_blocks: int,
        dim: int,
        flows: Optional[Iterable] = None,
        init_weight_scale: float = 1e-4,
        device: Union[str, int, torch.device] = 0,
        scale: float = 1.0,
    ):
        super().__init__()
        self.init_weight_scale = init_weight_scale
        self.x = None
        split_dim = dim // 2
        param_dims = [dim - split_dim, dim - split_dim]
        if flows is not None:
            self.flow = nn.ModuleList(flows)
        else:
            # hypernet = DenseNN(split_dim, [2 * dim], param_dims)
            self.flow = nn.ModuleList(
                [
                    AffineCoupling(
                        split_dim,
                        DenseNN(
                            split_dim,
                            [2 * dim],
                            param_dims,
                        ),
                    )
                    for _ in range(num_blocks)
                ],
            )
            self.init_params(self.parameters())

        even = [i for i in range(0, dim, 2)]
        odd = [i for i in range(1, dim, 2)]
        reverse_eo = [
            i // 2 if i % 2 == 0 else (i // 2 + len(even)) for i in range(dim)
        ]
        reverse_oe = [(i // 2 + len(odd)) if i % 2 == 0 else i // 2 for i in range(dim)]
        self.register_buffer("eo", torch.tensor(even + odd, dtype=torch.int64))
        self.register_buffer("oe", torch.tensor(odd + even, dtype=torch.int64))
        self.register_buffer(
            "reverse_eo",
            torch.tensor(reverse_eo, dtype=torch.int64),
        )
        self.register_buffer(
            "reverse_oe",
            torch.tensor(reverse_oe, dtype=torch.int64),
        )

        self.prior = MNormal(
            torch.zeros(dim).to(device), scale**2 * torch.eye(dim).to(device)
        )
        self.to(device)

    def init_params(self, params):
        # torch.nn.init.xavier_uniform_(params, gain=nn.init.calculate_gain('relu'))
        for p in params:
            if p.ndim == 2:
                torch.nn.init.sparse_(p, sparsity=0.3, std=self.init_weight_scale)
        # torch.nn.init.normal_(params, 0, self.init_weight_scale)

    def to(self, *args, **kwargs):
        """
        overloads to method to make sure the manually registered buffers are sent to device
        """
        self = super().to(*args, **kwargs)
        self.eo = self.eo.to(*args, **kwargs)
        self.oo = self.oe.to(*args, **kwargs)
        self.reverse_eo = self.reverse_eo.to(*args, **kwargs)
        self.reverse_oe = self.reverse_oe.to(*args, **kwargs)
        return self

    def permute(self, z, i, reverse=False):
        if not reverse:
            if i % 2 == 0:
                z = torch.index_select(z, -1, self.eo)
            else:
                z = torch.index_select(z, -1, self.oe)
        else:
            if i % 2 == 0:
                z = torch.index_select(z, -1, self.reverse_eo)
            else:
                z = torch.index_select(z, -1, self.reverse_oe)
        return z

    def forward(self, x):
        log_jacob = torch.zeros_like(x[..., 0], dtype=torch.float32)
        for i, current_flow in enumerate(self.flow):
            x = self.permute(x, i)
            z = current_flow(x)
            log_jacob += current_flow.log_abs_det_jacobian(x, z)  # z, z_new)
            z = self.permute(z, i, reverse=True)
            x = z
        return z, log_jacob

    def inverse(self, z):
        log_jacob_inv = torch.zeros_like(z[..., 0], dtype=torch.float32)
        n = len(self.flow) - 1
        for i, current_flow in enumerate(self.flow[::-1]):
            z = self.permute(z, n - i)
            x = current_flow._inverse(z)
            log_jacob_inv -= current_flow.log_abs_det_jacobian(x, z)
            x = self.permute(x, n - i, reverse=True)
            z = x
        return x, log_jacob_inv.reshape(z.shape[:-1])

    def log_prob(self, x):
        if self.x is not None and torch.equal(self.x, x):
            z, logp = self.z, self.log_jacob
        else:
            z, logp = self.forward(x)
        return self.prior.log_prob(z) + logp

    # def log_prob(self, x):
    #     z, logp = self.inverse(x)
    #     return self.prior.log_prob(z) + logp

    def sample(self, shape):
        z = self.prior.sample(shape)
        x, log_jacob_inv = self.inverse(z)
        self.log_jacob = -log_jacob_inv
        self.x = x
        self.z = z
        return x



from abc import ABC, abstractmethod
from typing import Callable, Dict, Optional

import torch


class Distribution(ABC):
    """Abstract class for distribution"""

    @abstractmethod
    def log_prob(self, z: torch.FloatTensor) -> torch.FloatTensor:
        """Computes log probability of input z"""
        raise NotImplementedError


class DistributionRegistry:
    registry: Dict = {}

    @classmethod
    def register(cls, name: Optional[str] = None) -> Callable:
        def inner_wrapper(wrapped_class: Distribution) -> Callable:
            if name is None:
                name_ = wrapped_class.__name__
            else:
                name_ = name
            cls.registry[name_] = wrapped_class
            return wrapped_class

        return inner_wrapper

    @classmethod
    def create(cls, name: str, **kwargs) -> Distribution:
        exec_class = cls.registry[name]
        executor = exec_class(**kwargs)
        return executor


@DistributionRegistry.register()
class PriorTarget(Distribution):
    def __init__(
        self,
        gan,
    ):
        self.gan = gan
        self.proposal = gan.prior

    def log_prob(self, z: torch.FloatTensor, **kwargs) -> torch.FloatTensor:
        return self.proposal.log_prob(z)

    def project(self, z):
        return self.proposal.project(z)


@DistributionRegistry.register()
class DiscriminatorTarget(Distribution):
    def __init__(self, gan, batch_size: Optional[int] = None):
        self.gan = gan
        self.proposal = gan.prior
        self.batch_size = batch_size
        self.device = next(self.gan.gen.parameters()).device

    def log_prob(self, z: torch.FloatTensor, **kwargs) -> torch.FloatTensor:
        init_shape = z.shape
        z = z.reshape(-1, init_shape[-1])
        batch_size = kwargs.get("batch_size", self.batch_size or len(z))
        log_prob = torch.empty((0,), device=self.device)
        for chunk_id, chunk in enumerate(torch.split(z, batch_size)):
            if "x" in kwargs:
                x = kwargs["x"][chunk_id * batch_size : (chunk_id + 1) * batch_size].to(
                    self.device
                )
            else:
                x = self.gan.gen(chunk.to(self.device))
            dgz = self.gan.dis(x).squeeze()
            logp_z = self.proposal.log_prob(chunk)
            log_prob = torch.cat([log_prob, (logp_z + dgz) / 1.0])
        return log_prob.reshape(init_shape[:-1])

    def project(self, z):
        return self.proposal.project(z)


@DistributionRegistry.register()
class CondTarget(Distribution):
    def __init__(
        self,
        gan,
        data_batch: Optional[torch.FloatTensor] = None,
        batch_size: Optional[int] = None,
    ):
        self.gan = gan
        self.proposal = gan.prior
        self.data_batch = data_batch
        self.batch_size = batch_size

    def log_prob(
        self,
        z: torch.FloatTensor,
        data_batch: Optional[torch.FloatTensor] = None,
        **kwargs,
    ) -> torch.FloatTensor:
        batch_size = kwargs.get("batch_size", self.batch_size or len(z))
        data_batch = data_batch if data_batch is not None else self.data_batch
        log_prob = torch.empty((0,), device=z.device)
        for chunk, data_chunk in zip(
            torch.split(z, batch_size), torch.split(data_batch, batch_size)
        ):
            logp_xz = (
                -torch.norm(
                    (self.gan.gen(chunk) - data_chunk.to(chunk.device)).reshape(
                        len(chunk), -1
                    ),
                    dim=1,
                )
                ** 2
                / 2
            )
            logp_z = self.proposal.log_prob(chunk)
            if logp_xz.shape != logp_z.shape:
                raise Exception
            log_prob = torch.cat([log_prob, (logp_z + logp_xz) / 1.0])

        return log_prob

    def project(self, z):
        return self.proposal.project(z)



""" 
Code is borrowed from https://github.com/bioinf-jku/TTUR repository

Calculates the Frechet Inception Distance (FID) to evalulate GANs.

The FID metric calculates the distance between two distributions of images.
Typically, we have summary statistics (mean & covariance matrix) of one
of these distributions, while the 2nd distribution is given by a GAN.

When run as a stand-alone program, it compares the distribution of
images that are stored as PNG/JPEG at a specified location with a
distribution given by summary statistics (in pickle format).

The FID is calculated by assuming that X_1 and X_2 are the activations of
the pool_3 layer of the inception net for generated samples and real world
samples respectivly.

See --help to see further details.
"""

from __future__ import absolute_import, division, print_function

import os
import warnings

import numpy as np


os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

warnings.filterwarnings("ignore")
import tensorflow.compat.v1 as tf


tf.disable_v2_behavior()

import logging
import pathlib

from imageio import imread
from scipy import linalg
from tqdm import trange


tf.logging.set_verbosity(logging.ERROR)


class InvalidFIDException(Exception):
    pass


def create_inception_graph(pth):
    """Creates a graph from saved GraphDef file."""
    # Creates graph from saved graph_def.pb.
    with tf.io.gfile.GFile(pth, "rb") as f:
        graph_def = tf.compat.v1.GraphDef()
        graph_def.ParseFromString(f.read())
        _ = tf.import_graph_def(graph_def, name="FID_Inception_Net")


# -------------------------------------------------------------------------------


# code for handling inception net derived from
#   https://github.com/openai/improved-gan/blob/master/inception_score/model.py
def _get_inception_layer(sess):
    """Prepares inception net for batched usage and returns pool_3 layer."""
    layername = "FID_Inception_Net/pool_3:0"
    pool3 = sess.graph.get_tensor_by_name(layername)
    ops = pool3.graph.get_operations()
    for op_idx, op in enumerate(ops):
        for o in op.outputs:
            shape = o.get_shape()
            if shape._dims is not None:
                # shape = [s.value for s in shape] TF 1.x
                shape = [s for s in shape]  # TF 2.x
                new_shape = []
                for j, s in enumerate(shape):
                    if s == 1 and j == 0:
                        new_shape.append(None)
                    else:
                        new_shape.append(s)
                o.__dict__["_shape_val"] = tf.TensorShape(new_shape)
    return pool3


# -------------------------------------------------------------------------------


def get_activations(images, sess, batch_size=100, verbose=True):  # False):
    """Calculates the activations of the pool_3 layer for all images.

    Params:
    -- images      : Numpy array of dimension (n_images, hi, wi, 3). The values
                     must lie between 0 and 256.
    -- sess        : current session
    -- batch_size  : the images numpy array is split into batches with batch size
                     batch_size. A reasonable batch size depends on the disposable hardware.
    -- verbose    : If set to True and parameter out_step is given, the number of calculated
                     batches is reported.
    Returns:
    -- A numpy array of dimension (num images, 2048) that contains the
       activations of the given tensor when feeding inception with the query tensor.
    """
    inception_layer = _get_inception_layer(sess)
    n_images = images.shape[0]
    if batch_size > n_images:
        print(
            "warning: batch size is bigger than the data size. setting batch size to data size"
        )
        batch_size = n_images
    n_batches = n_images // batch_size  # drops the last batch if < batch_size
    pred_arr = np.empty((n_batches * batch_size, 2048))
    if verbose:
        pbar = trange
    else:
        pbar = range
    for i in pbar(n_batches):
        if verbose:
            print(
                "\rPropagating batch %d/%d" % (i + 1, n_batches),
                end="",
                flush=True,
            )
        start = i * batch_size

        if start + batch_size < n_images:
            end = start + batch_size
        else:
            end = n_images

        batch = images[start:end]
        pred = sess.run(inception_layer, {"FID_Inception_Net/ExpandDims:0": batch})
        pred_arr[start:end] = pred.reshape(batch.shape[0], -1)
    if verbose:
        print(" done")
    return pred_arr


# -------------------------------------------------------------------------------


def calculate_frechet_distance(mu1, sigma1, mu2, sigma2, eps=1e-6):
    """Numpy implementation of the Frechet Distance.
    The Frechet distance between two multivariate Gaussians X_1 ~ N(mu_1, C_1)
    and X_2 ~ N(mu_2, C_2) is
            d^2 = ||mu_1 - mu_2||^2 + Tr(C_1 + C_2 - 2*sqrt(C_1*C_2)).

    Stable version by Dougal J. Sutherland.

    Params:
    -- mu1 : Numpy array containing the activations of the pool_3 layer of the
             inception net ( like returned by the function 'get_predictions')
             for generated samples.
    -- mu2   : The sample mean over activations of the pool_3 layer, precalcualted
               on an representive data set.
    -- sigma1: The covariance matrix over activations of the pool_3 layer for
               generated samples.
    -- sigma2: The covariance matrix over activations of the pool_3 layer,
               precalcualted on an representive data set.

    Returns:
    --   : The Frechet Distance.
    """

    mu1 = np.atleast_1d(mu1)
    mu2 = np.atleast_1d(mu2)

    sigma1 = np.atleast_2d(sigma1)
    sigma2 = np.atleast_2d(sigma2)

    assert (
        mu1.shape == mu2.shape
    ), "Training and test mean vectors have different lengths"
    assert (
        sigma1.shape == sigma2.shape
    ), "Training and test covariances have different dimensions"

    diff = mu1 - mu2

    # product might be almost singular
    covmean, _ = linalg.sqrtm(sigma1.dot(sigma2), disp=False)
    if not np.isfinite(covmean).all():
        msg = (
            "fid calculation produces singular product; adding %s to diagonal of cov estimates"
            % eps
        )
        warnings.warn(msg)
        offset = np.eye(sigma1.shape[0]) * eps
        covmean = linalg.sqrtm((sigma1 + offset).dot(sigma2 + offset))

    # numerical error might give slight imaginary component
    if np.iscomplexobj(covmean):
        if not np.allclose(np.diagonal(covmean).imag, 0, atol=1e-3):
            m = np.max(np.abs(covmean.imag))
            raise ValueError("Imaginary component {}".format(m))
        covmean = covmean.real

    tr_covmean = np.trace(covmean)

    return diff.dot(diff) + np.trace(sigma1) + np.trace(sigma2) - 2 * tr_covmean


# -------------------------------------------------------------------------------


def calculate_activation_statistics(images, sess, batch_size=50, verbose=False):
    """Calculation of the statistics used by the FID.
    Params:
    -- images      : Numpy array of dimension (n_images, hi, wi, 3). The values
                     must lie between 0 and 255.
    -- sess        : current session
    -- batch_size  : the images numpy array is split into batches with batch size
                     batch_size. A reasonable batch size depends on the available hardware.
    -- verbose     : If set to True and parameter out_step is given, the number of calculated
                     batches is reported.
    Returns:
    -- mu    : The mean over samples of the activations of the pool_3 layer of
               the incption model.
    -- sigma : The covariance matrix of the activations of the pool_3 layer of
               the incption model.
    """
    act = get_activations(images, sess, batch_size, verbose)
    mu = np.mean(act, axis=0)
    sigma = np.cov(act, rowvar=False)
    return mu, sigma


# ------------------
# The following methods are implemented to obtain a batched version of the activations.
# This has the advantage to reduce memory requirements, at the cost of slightly reduced efficiency.
# - Pyrestone
# ------------------


def load_image_batch(files):
    """Convenience method for batch-loading images
    Params:
    -- files    : list of paths to image files. Images need to have same dimensions for all files.
    Returns:
    -- A numpy array of dimensions (num_images,hi, wi, 3) representing the image pixel values.
    """
    return np.array([imread(str(fn)).astype(np.float32) for fn in files])


def get_activations_from_files(files, sess, batch_size=50, verbose=False):
    """Calculates the activations of the pool_3 layer for all images.

    Params:
    -- files      : list of paths to image files. Images need to have same dimensions for all files.
    -- sess        : current session
    -- batch_size  : the images numpy array is split into batches with batch size
                     batch_size. A reasonable batch size depends on the disposable hardware.
    -- verbose    : If set to True and parameter out_step is given, the number of calculated
                     batches is reported.
    Returns:
    -- A numpy array of dimension (num images, 2048) that contains the
       activations of the given tensor when feeding inception with the query tensor.
    """
    inception_layer = _get_inception_layer(sess)
    n_imgs = len(files)
    if batch_size > n_imgs:
        print(
            "warning: batch size is bigger than the data size. setting batch size to data size"
        )
        batch_size = n_imgs
    n_batches = n_imgs // batch_size + 1
    pred_arr = np.empty((n_imgs, 2048))
    for i in range(n_batches):
        if verbose:
            print(
                "\rPropagating batch %d/%d" % (i + 1, n_batches),
                end="",
                flush=True,
            )
        start = i * batch_size
        if start + batch_size < n_imgs:
            end = start + batch_size
        else:
            end = n_imgs

        batch = load_image_batch(files[start:end])
        pred = sess.run(inception_layer, {"FID_Inception_Net/ExpandDims:0": batch})
        pred_arr[start:end] = pred.reshape(batch_size, -1)
        del batch  # clean up memory
    if verbose:
        print(" done")
    return pred_arr


def calculate_activation_statistics_from_files(
    files, sess, batch_size=50, verbose=False
):
    """Calculation of the statistics used by the FID.
    Params:
    -- files      : list of paths to image files. Images need to have same dimensions for all files.
    -- sess        : current session
    -- batch_size  : the images numpy array is split into batches with batch size
                     batch_size. A reasonable batch size depends on the available hardware.
    -- verbose     : If set to True and parameter out_step is given, the number of calculated
                     batches is reported.
    Returns:
    -- mu    : The mean over samples of the activations of the pool_3 layer of
               the incption model.
    -- sigma : The covariance matrix of the activations of the pool_3 layer of
               the incption model.
    """
    act = get_activations_from_files(files, sess, batch_size, verbose)
    mu = np.mean(act, axis=0)
    sigma = np.cov(act, rowvar=False)
    return mu, sigma


# -------------------------------------------------------------------------------


# -------------------------------------------------------------------------------
# The following functions aren't needed for calculating the FID
# they're just here to make this module work as a stand-alone script
# for calculating FID scores
# -------------------------------------------------------------------------------
def check_or_download_inception(inception_path):
    """Checks if the path to the inception file is valid, or downloads
    the file if it is not present."""
    INCEPTION_URL = (
        "http://download.tensorflow.org/models/image/imagenet/inception-2015-12-05.tgz"
    )
    if inception_path is None:
        inception_path = "/tmp"
    inception_path = pathlib.Path(inception_path)
    model_file = inception_path / "classify_image_graph_def.pb"
    if not model_file.exists():
        print("Downloading Inception model")
        import tarfile
        from urllib import request

        fn, _ = request.urlretrieve(INCEPTION_URL)
        with tarfile.open(fn, mode="r") as f:
            f.extract("classify_image_graph_def.pb", str(model_file.parent))
    return str(model_file)


def _handle_path(path, sess, low_profile=False):
    if path.endswith(".npz"):
        f = np.load(path)
        m, s = f["mu"][:], f["sigma"][:]
        f.close()
    else:
        path = pathlib.Path(path)
        imgs = np.load(path).transpose(0, 2, 3, 1) * 255
        m, s = calculate_activation_statistics(imgs, sess)
        del imgs
        # files = list(path.glob('*.jpg')) + list(path.glob('*.png'))
        # if low_profile:
        #     m, s = calculate_activation_statistics_from_files(files, sess)
        # else:
        #     x = np.array([imread(str(fn)).astype(np.float32) for fn in files])
        #     m, s = calculate_activation_statistics(x, sess)
        #     del x #clean up memory
    return m, s


def calculate_fid_given_paths(paths, inception_path, low_profile=False):
    """Calculates the FID of two paths."""
    inception_path = check_or_download_inception(inception_path)

    for p in paths:
        if not os.path.exists(p):
            raise RuntimeError("Invalid path: %s" % p)

    config = tf.ConfigProto()
    config.gpu_options.allow_growth = True
    create_inception_graph(str(inception_path))
    with tf.Session(config=config) as sess:
        sess.run(tf.global_variables_initializer())
        m1, s1 = _handle_path(paths[0], sess, low_profile=low_profile)
        m2, s2 = _handle_path(paths[1], sess, low_profile=low_profile)
        fid_value = calculate_frechet_distance(m1, s1, m2, s2)
        return fid_value


def get_activation_statistics(path, inception_path):
    inception_path = check_or_download_inception(inception_path)

    config = tf.ConfigProto()
    config.gpu_options.allow_growth = True
    create_inception_graph(str(inception_path))
    with tf.Session(config=config) as sess:
        sess.run(tf.global_variables_initializer())
        m, s = _handle_path(path, sess)
    return m, s


if __name__ == "__main__":
    from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser

    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        "path",
        type=str,
        nargs=2,
        help="Path to the generated images or to .npz statistic files",
    )
    parser.add_argument(
        "-i",
        "--inception",
        type=str,
        default=None,
        help="Path to Inception model (will be downloaded if not provided)",
    )
    parser.add_argument(
        "--gpu",
        default="",
        type=str,
        help="GPU to use (leave blank for CPU only)",
    )
    parser.add_argument(
        "--lowprofile",
        action="store_true",
        help="Keep only one batch of images in memory at a time. This reduces memory footprint, but may decrease speed slightly.",
    )
    args = parser.parse_args()
    os.environ["CUDA_VISIBLE_DEVICES"] = args.gpu
    fid_value = calculate_fid_given_paths(
        args.path, args.inception, low_profile=args.lowprofile
    )
    print("FID: ", fid_value)



from typing import Tuple

from torch_mimicry.nets import dcgan, sngan

from .base import BaseDiscriminator, BaseGenerator, ModelRegistry


@ModelRegistry.register()
class MMCDCGenerator(BaseGenerator):
    def __init__(
        self, mean: Tuple[float, float, float], std: Tuple[float, float, float]
    ):
        super().__init__(mean, std)
        self.gen = dcgan.DCGANGenerator32()
        self.z_dim = self.gen.nz

    def load_state_dict(self, state_dict, strict: bool = True):
        return self.gen.load_state_dict(state_dict["model_state_dict"], strict=strict)

    def forward(self, z):
        return self.gen(z)


@ModelRegistry.register()
class MMCDCDiscriminator(BaseDiscriminator):
    def __init__(
        self,
        mean: Tuple[float, float, float],
        std: Tuple[float, float, float],
        output_layer="identity",
    ):
        super().__init__(mean, std, output_layer)
        self.dis = dcgan.DCGANDiscriminator32()

    @property
    def penult_layer(self):
        return self.dis.activation

    def load_state_dict(self, state_dict, strict: bool = True):
        return self.dis.load_state_dict(state_dict["model_state_dict"], strict=strict)

    def forward(self, x):
        return self.dis(x)


@ModelRegistry.register()
class MMCSNGenerator(BaseGenerator):
    def __init__(
        self, mean: Tuple[float, float, float], std: Tuple[float, float, float]
    ):
        super().__init__(mean, std)
        self.gen = sngan.SNGANGenerator32()
        self.z_dim = self.gen.nz

    def load_state_dict(self, state_dict, strict: bool = True):
        return self.gen.load_state_dict(state_dict["model_state_dict"], strict=strict)

    def forward(self, z):
        return self.gen(z)


@ModelRegistry.register()
class MMCSNDiscriminator(BaseDiscriminator):
    def __init__(
        self,
        mean: Tuple[float, float, float],
        std: Tuple[float, float, float],
        output_layer="identity",
    ):
        super().__init__(mean, std, output_layer)
        self.dis = sngan.SNGANDiscriminator32()

    @property
    def penult_layer(self):
        return self.dis.activation

    def load_state_dict(self, state_dict, strict: bool = True):
        return self.dis.load_state_dict(state_dict["model_state_dict"], strict=strict)

    def forward(self, x):
        return self.dis(x)



from typing import Union

import torch
from torch import nn
from torch.distributions import MultivariateNormal as MNormal


class MinimalRNVP(nn.Module):
    def __init__(
        self,
        dim: int,
        device: Union[str, int, torch.device],
        hidden: int = 32,
        num_blocks: int = 4,
        init_weight_scale: float = 1e-4,
        scale: float = 1.0,
    ):
        self.init_weight_scale = init_weight_scale
        device = torch.device(device)
        super().__init__()

        self.prior = MNormal(
            torch.zeros(dim, requires_grad=False).to(device),
            scale**2 * torch.eye(dim, requires_grad=False).to(device),
        )

        masks = num_blocks * [
            [i % 2 for i in range(dim)],
            [(i + 1) % 2 for i in range(dim)],
        ]
        masks = torch.FloatTensor(masks)
        self.masks = nn.Parameter(masks, requires_grad=False)

        self.t = torch.nn.ModuleList(
            [
                nn.Sequential(
                    nn.Linear(dim, hidden),
                    nn.LeakyReLU(),
                    nn.Linear(hidden, hidden),
                    nn.LeakyReLU(),
                    nn.Linear(hidden, dim),
                )
                for _ in range(2 * num_blocks)
            ]
        )
        self.s = torch.nn.ModuleList(
            [
                nn.Sequential(
                    nn.Linear(dim, hidden),
                    nn.LeakyReLU(),
                    nn.Linear(hidden, hidden),
                    nn.LeakyReLU(),
                    nn.Linear(hidden, dim),
                    nn.Tanh(),
                )
                for _ in range(2 * num_blocks)
            ]
        )

        self.to(device)
        self.init_params([p for n, p in self.named_parameters() if "weight" in n])

    def init_params(self, params):
        # torch.nn.init.xavier_uniform_(params, gain=nn.init.calculate_gain('relu')) #45
        for p in params:
            torch.nn.init.sparse_(p, sparsity=0.3, std=self.init_weight_scale)
            # torch.nn.init.normal_(p, 0, self.init_weight_scale)

    def inverse_flatten(self, z):
        log_det_J_inv, x = z.new_zeros(z.shape[0]), z
        for i in range(len(self.t)):
            x_ = x * self.masks[i]
            s = self.s[i](x_) * (1 - self.masks[i])
            t = self.t[i](x_) * (1 - self.masks[i])
            x = x_ + (1 - self.masks[i]) * (x * torch.exp(s) + t)

            log_det_J_inv += s.sum(dim=1)
        return x, log_det_J_inv

    def inverse(self, z):
        dim = z.shape[-1]
        first_dims = z.shape[:-1]

        x, log_det_J_inv = self.inverse_flatten(z.reshape(-1, dim))
        x = x.reshape(*z.shape)
        log_det_J_inv = log_det_J_inv.reshape(*first_dims)

        return x, log_det_J_inv

    def forward_flatten(self, x):
        log_det_J, z = x.new_zeros(x.shape[0]), x
        for i in reversed(range(len(self.t))):
            z_ = self.masks[i] * z
            s = self.s[i](z_) * (1 - self.masks[i])
            t = self.t[i](z_) * (1 - self.masks[i])
            z = (1 - self.masks[i]) * (z - t) * torch.exp(-s) + z_
            log_det_J -= s.sum(dim=1)
        return z, log_det_J

    def forward(self, x):
        dim = x.shape[-1]
        first_dims = x.shape[:-1]

        z, log_det_J = self.forward_flatten(x.reshape(-1, dim))
        z = z.reshape(*x.shape)
        log_det_J = log_det_J.reshape(*first_dims)

        return z, log_det_J

    def log_prob(self, x):
        z, logp = self.forward(x)
        return self.prior.log_prob(z) + logp

    def sample(self, shape):
        z = self.prior.sample(shape).detach()
        x, _ = self.inverse(z)
        # self.log_det_J_inv = log_det_J_inv
        return x



from typing import Optional, Tuple

import torch
from torch import nn
from torchvision import transforms


class ModelRegistry:
    registry = {}

    @classmethod
    def register(cls, name: Optional[str] = None) -> nn.Module:
        def inner_wrapper(wrapped_class: nn.Module) -> nn.Module:
            if name is None:
                name_ = wrapped_class.__name__
            else:
                name_ = name
            cls.registry[name_] = wrapped_class
            return wrapped_class

        return inner_wrapper

    @classmethod
    def create(cls, name: str, **kwargs) -> nn.Module:
        model = cls.registry[name]
        model = model(**kwargs)
        return model


class NormalizeInverse(transforms.Normalize):
    """
    Undoes the normalization and returns the reconstructed images in the input domain.
    """

    def __init__(self, mean, std):
        mean = torch.as_tensor(mean)
        std = torch.as_tensor(std)
        std_inv = 1 / (std + 1e-10)
        mean_inv = -mean * std_inv
        super().__init__(mean=mean_inv, std=std_inv)

    def __call__(self, tensor):
        return super().__call__(tensor.clone())


class BaseDiscriminator(nn.Module):
    def __init__(
        self,
        mean: Tuple[float, float, float],
        std: Tuple[float, float, float],
        output_layer: str,
    ):
        super().__init__()
        self.transform = transforms.Normalize(mean, std)
        if output_layer == "sigmoid":
            self.output_layer = nn.Sigmoid()
        else:
            self.output_layer = nn.Identity()
        # self.input = None
        # self.output = None
        # self.register_forward_hook(forward_memory_hook)
        # self.register_backward_hook(backward_hook)

    def get_label(self) -> torch.LongTensor:
        return self.label.data.long()


class BaseGenerator(nn.Module):
    def __init__(
        self, mean: Tuple[float, float, float], std: Tuple[float, float, float]
    ):
        super().__init__()
        self.inverse_transform = transforms.Compose(
            [
                NormalizeInverse(mean, std),
                transforms.Lambda(lambda x: torch.clip(x, 0, 1)),
            ]
        )
        # self.input = None
        # self.output = None
        # self.register_forward_hook(forward_memory_hook)

    def sample_label(self, *args, **kwargs):
        return None

    def get_label(self) -> torch.LongTensor:
        return self.label.data.long()


class MemoryModel(nn.Module):
    def __init__(self, module: nn.Module):
        super().__init__()
        self.module = module
        self.input = None
        self.output = None

    def forward(self, input):
        if self.input is not None and torch.equal(self.input, input):
            output = self.output
        else:
            output = self.module.forward(input)
            self.output = output
            self.input = input
        return output
